Package: DICErClust
Title: Deep Significance Clustering for Clinical Risk Stratification
Version: 0.1.2
Authors@R: 
    c(person(given="Sarah", family="Ayton", , "saa7050@med.cornell.edu", role = c("aut", "cre"), comment = c(ORCID = "0000-0001-5247-9912")),
    person(given="Yiye", family="Zhang", , "yiz2014@med.cornell.edu", role = c("aut"), comment = c(ORCID = "0000-0003-3494-2699")))
Description: We provide an R implementation of Deep Significance Clustering (DICE), a self-supervised learning framework designed to identify clinically meaningful and risk-stratified patient subgroups from electronic health record (EHR) data. DICE jointly optimizes deep representation learning, clustering, and outcome prediction while enforcing statistical significance between predicted outcomes and cluster membership. This integrated optimization produces subgroups that are both clinically coherent and predictive, addressing a gap where traditional unsupervised clustering methods and supervised risk prediction models alone may fail to generate actionable clinical groupings. See Huang et al. (2021) <doi:10.1093/jamia/ocab203>.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: argparser, ggplot2, torch, pROC
Suggests: knitr, rmarkdown, scales
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-21 22:05:42 UTC; sarahayton
Author: Sarah Ayton [aut, cre] (ORCID: <https://orcid.org/0000-0001-5247-9912>),
  Yiye Zhang [aut] (ORCID: <https://orcid.org/0000-0003-3494-2699>)
Maintainer: Sarah Ayton <saa7050@med.cornell.edu>
Config/roxygen2/version: 8.0.0
Repository: CRAN
Date/Publication: 2026-05-28 12:50:09 UTC
Built: R 4.5.2; ; 2026-05-28 15:53:52 UTC; unix

Package: DICErClust
Title: Deep Significance Clustering for Clinical Risk Stratification
Version: 0.1.2
Authors@R: 
    c(person(given="Sarah", family="Ayton", , "saa7050@med.cornell.edu", role = c("aut", "cre"), comment = c(ORCID = "0000-0001-5247-9912")),
    person(given="Yiye", family="Zhang", , "yiz2014@med.cornell.edu", role = c("aut"), comment = c(ORCID = "0000-0003-3494-2699")))
Description: We provide an R implementation of Deep Significance Clustering (DICE), a self-supervised learning framework designed to identify clinically meaningful and risk-stratified patient subgroups from electronic health record (EHR) data. DICE jointly optimizes deep representation learning, clustering, and outcome prediction while enforcing statistical significance between predicted outcomes and cluster membership. This integrated optimization produces subgroups that are both clinically coherent and predictive, addressing a gap where traditional unsupervised clustering methods and supervised risk prediction models alone may fail to generate actionable clinical groupings. See Huang et al. (2021) <doi:10.1093/jamia/ocab203>.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: argparser, ggplot2, torch, pROC
Suggests: knitr, rmarkdown, scales
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-21 22:05:42 UTC; sarahayton
Author: Sarah Ayton [aut, cre] (ORCID: <https://orcid.org/0000-0001-5247-9912>),
  Yiye Zhang [aut] (ORCID: <https://orcid.org/0000-0003-3494-2699>)
Maintainer: Sarah Ayton <saa7050@med.cornell.edu>
Config/roxygen2/version: 8.0.0
Repository: CRAN
Date/Publication: 2026-05-28 12:50:09 UTC
Built: R 4.6.0; ; 2026-05-28 16:36:41 UTC; unix

# Generated by roxygen2: do not edit by hand

export(DICEr)
export(DecoderRNN)
export(EncoderRNN)
export(analysis_p_value_related)
export(analysis_p_value_related_onlyc)
export(change_label_from_highratio_to_lowratio)
export(func_analysis_test_error_D0406)
export(main)
export(model_2)
export(p_value_calculate)
export(parse_args)
export(test_AE)
export(update_M)
export(update_testset_R_C_M_K)
import(argparser)
import(ggplot2)
import(stats)
import(torch)

# DICErClust

**Deep Significance Clustering for Clinical Risk Stratification**

DICErClust provides an R implementation of the DICE (Deep Significance
Clustering) algorithm, a self-supervised deep learning framework that
identifies clinically meaningful, risk-stratified patient subgroups from
electronic health record (EHR) data.

DICE jointly optimises four objectives — autoencoder reconstruction, cluster
cohesion, binary outcome prediction, and a likelihood-ratio test (LRT)
significance penalty — so that the discovered clusters are both data-driven and
statistically validated.

> Huang Y, Du C, Zhu F, *et al.* (2021). Self-supervised deep clustering of
> patient subgroups for heart failure with preserved ejection fraction.
> *J Am Med Inform Assoc*, **28**, 2394–2403.
> doi:[10.1093/jamia/ocab203](https://doi.org/10.1093/jamia/ocab203)

---

## Installation

DICErClust is distributed as a source tarball.  Install it once with
`install.packages()`, then load it like any other R package.

```r
install.packages(
  "/path/to/DICErClust_0.1.1.tar.gz",
  repos = NULL, type = "source"
)
```

DICErClust depends on the **torch** package for R.  Install torch and its
runtime binaries with:

```r
install.packages("torch")
torch::install_torch()   # downloads the LibTorch runtime (~500 MB, once only)
```

---

## Minimal working example

```r
library(DICErClust)

## ── 1. Prepare data ─────────────────────────────────────────────────────────
## DICEr() reads RDS files, each a length-3 list:
##   [[1]] data_x  — numeric matrix n × p (continuous features, LSTM input)
##   [[2]] data_v  — numeric matrix n × q (binary demographics, outcome head)
##   [[3]] data_y  — integer vector 0/1 outcome

set.seed(1); n <- 150L; p <- 5L; q <- 3L
data_dir <- file.path(tempdir(), "dice_demo")
dir.create(data_dir, showWarnings = FALSE)

saveRDS(list(matrix(runif(n * p), n, p),
             matrix(as.numeric(rbinom(n * q, 1, 0.5)), n, q),
             rbinom(n, 1, 0.3)),
        file.path(data_dir, "train.rds"))
saveRDS(list(matrix(runif(50L * p), 50L, p),
             matrix(as.numeric(rbinom(50L * q, 1, 0.5)), 50L, q),
             rbinom(50L, 1, 0.3)),
        file.path(data_dir, "test.rds"))

## ── 2. Configure and train ──────────────────────────────────────────────────
args <- list(
  seed = 1L, input_path = data_dir,
  filename_train = "train.rds", filename_test = "test.rds",
  n_input_fea = p, n_hidden_fea = 3L,
  lstm_layer = 1L, lstm_dropout = 0.0, K_clusters = 2L,
  n_dummy_demov_fea = q, cuda = FALSE, lr = 1e-4,
  init_AE_epoch = 5L, iter = 20L, epoch_in_iter = 2L,
  lambda_AE = 1.0, lambda_classifier = 1.0,
  lambda_outcome = 1.0, lambda_p_value = 1.0
)

old_wd <- setwd(tempdir())
DICEr(args)          # writes output to hn_3_K_2/part2_AE_nhidden_3/
setwd(old_wd)

## ── 3. Load best checkpoint ─────────────────────────────────────────────────
part2_dir <- file.path(tempdir(), "hn_3_K_2", "part2_AE_nhidden_3")
res_train <- readRDS(file.path(part2_dir, "data_train_iter.rds"))
res_test  <- readRDS(file.path(part2_dir, "data_test_iter.rds"))

## Use res_test$pred_C (not $C) for test-set cluster labels
table(res_test$pred_C)
```

---

## Vignettes

| Vignette | Description |
|---|---|
| [Introduction to DICErClust](vignettes/DICEr-introduction.Rmd) | Package overview, data format, quick start, hyperparameter guide |
| [Heart Failure Example](vignettes/heart-failure-example.Rmd) | Full end-to-end analysis on the UCI Heart Failure dataset (AUC = 0.823, χ² = 32.99, *p* < 0.001) |

To browse vignettes after installation:

```r
vignette("DICEr-introduction",      package = "DICErClust")
vignette("heart-failure-example",   package = "DICErClust")
```

---

## How it works

DICEr runs three sequential stages:

1. **Autoencoder warm-up** (`init_AE_epoch` epochs): an LSTM encoder–decoder
   is pre-trained on `data_x` to learn compact latent representations before
   any clustering begins.

2. **Joint optimisation** (`iter` iterations × `epoch_in_iter` epochs):
   k-means clustering alternates with gradient updates that minimise:

   ```
   L = λ_AE · L_recon  +  λ_clf · L_cluster  +  λ_out · L_outcome  +  λ_p · L_LRT
   ```

   where `L_LRT = 3.841 − G` penalises cluster configurations that do not
   reach χ²(1) significance at α = 0.05.

3. **Model selection**: the checkpoint with the lowest test negative
   log-likelihood that also satisfies LRT *p* < 0.05 is saved.

---

## Citation

```bibtex
@article{huang2021self,
  title   = {Self-supervised deep clustering of patient subgroups for
             heart failure with preserved ejection fraction},
  author  = {Huang, Yiye and Du, Changchun and Zhu, Fan and others},
  journal = {Journal of the American Medical Informatics Association},
  volume  = {28},
  pages   = {2394--2403},
  year    = {2021},
  doi     = {10.1093/jamia/ocab203}
}
```

---

## License

MIT © Sarah Ayton and Yiye Zhang, Weill Cornell Medicine

## ----setup, include = FALSE---------------------------------------------------
## All computational chunks are skipped on CRAN: this vignette downloads
## external data and runs DICEr() (~15-20 min on CPU), both of which are
## incompatible with CRAN's check environment.
## devtools::check() sets NOT_CRAN=true automatically for local builds.
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.width  = 5,
  fig.height = 4,
  eval       = identical(Sys.getenv("NOT_CRAN"), "true")
)

## ----load-pkg, eval = FALSE---------------------------------------------------
# ## Install from local tarball (run once):
# # install.packages(
# #   "/path/to/DICErClust_0.1.1.tar.gz",
# #   repos = NULL, type = "source"
# # )
# library(DICErClust)
# library(ggplot2)
# library(pROC)

## ----load-pkg-real, include = FALSE-------------------------------------------
# ## When building the vignette from within the package source tree we use
# ## devtools::load_all() so edits to the source are reflected immediately.
# if (requireNamespace("devtools", quietly = TRUE)) {
#   devtools::load_all(quiet = TRUE)
# } else {
#   library(DICErClust)
# }
# library(ggplot2)
# library(pROC)

## ----download-data------------------------------------------------------------
# hf_url  <- paste0(
#   "https://archive.ics.uci.edu/ml/",
#   "machine-learning-databases/00519/",
#   "heart_failure_clinical_records_dataset.csv"
# )
# hf_dest <- tempfile(fileext = ".csv")
# download.file(hf_url, hf_dest, quiet = TRUE)
# hf <- read.csv(hf_dest)
# 
# cat(sprintf("Rows: %d   Columns: %d\n", nrow(hf), ncol(hf)))
# print(table(DEATH_EVENT = hf$DEATH_EVENT))

## ----features-----------------------------------------------------------------
# ## Continuous lab features → LSTM encoder (data_x)
# x_cols <- c("age", "creatinine_phosphokinase", "ejection_fraction",
#             "platelets", "serum_creatinine", "serum_sodium", "time")
# 
# ## Binary demographic indicators → outcome head (data_v)
# v_cols <- c("anaemia", "diabetes", "high_blood_pressure", "sex", "smoking")
# 
# ## Min-max scale continuous features to [0, 1].
# ## Scaling prevents any single lab value from dominating the MSE
# ## reconstruction loss relative to others.
# scale_01 <- function(x) {
#   r <- range(x, na.rm = TRUE)
#   if (diff(r) == 0) return(x * 0)
#   (x - r[1]) / diff(r)
# }
# 
# X_x <- apply(as.matrix(hf[, x_cols]), 2, scale_01)  # 299 × 7, numeric
# X_v <- apply(as.matrix(hf[, v_cols]), 2, as.numeric) # 299 × 5, binary as float
# 
# ## Note: data_v *must* be stored as numeric (float), not integer.
# ## torch_tensor() infers dtype from R storage mode; integer columns produce
# ## int64 tensors that are incompatible with the float32 model weights.
# 
# cat(sprintf("data_x: %d × %d\ndata_v: %d × %d\n",
#             nrow(X_x), ncol(X_x), nrow(X_v), ncol(X_v)))
# 
# n_x <- ncol(X_x)  # 7  continuous predictors
# n_v <- ncol(X_v)  # 5  binary demographics
# outcome <- hf$DEATH_EVENT

## ----split--------------------------------------------------------------------
# set.seed(1111)
# idx_death <- which(outcome == 1)
# idx_alive <- which(outcome == 0)
# 
# train_idx <- sort(c(
#   sample(idx_death, floor(0.70 * length(idx_death))),
#   sample(idx_alive, floor(0.70 * length(idx_alive)))
# ))
# test_idx <- setdiff(seq_len(nrow(hf)), train_idx)
# 
# cat(sprintf("Train: %d patients  (deaths: %d, %.0f%%)\n",
#             length(train_idx), sum(outcome[train_idx]),
#             100 * mean(outcome[train_idx])))
# cat(sprintf("Test : %d patients  (deaths: %d, %.0f%%)\n",
#             length(test_idx),  sum(outcome[test_idx]),
#             100 * mean(outcome[test_idx])))

## ----save-rds-----------------------------------------------------------------
# data_dir <- file.path(tempdir(), "dice_hf")
# dir.create(data_dir, showWarnings = FALSE)
# 
# saveRDS(
#   list(X_x[train_idx, ], X_v[train_idx, ], as.integer(outcome[train_idx])),
#   file.path(data_dir, "hf_train.rds")
# )
# saveRDS(
#   list(X_x[test_idx, ], X_v[test_idx, ], as.integer(outcome[test_idx])),
#   file.path(data_dir, "hf_test.rds")
# )

## ----configure----------------------------------------------------------------
# args <- list(
#   seed              = 1111,          # reproducibility seed
#   input_path        = data_dir,      # directory containing RDS files
#   filename_train    = "hf_train.rds",
#   filename_test     = "hf_test.rds",
# 
#   ## ── Architecture ──────────────────────────────────────────
#   n_input_fea       = n_x,          # 7 continuous LSTM input features
#   n_hidden_fea      = 4,            # LSTM latent dimension (7 → 4)
#   lstm_layer        = 1,            # single LSTM layer
#   lstm_dropout      = 0.0,          # no dropout (small dataset)
#   K_clusters        = 2,            # binary risk partition: high vs. low
# 
#   ## ── Auxiliary features ────────────────────────────────────
#   n_dummy_demov_fea = n_v,          # 5 binary demographic covariates
# 
#   ## ── Hardware ──────────────────────────────────────────────
#   cuda              = FALSE,        # set TRUE for GPU acceleration
# 
#   ## ── Optimiser ─────────────────────────────────────────────
#   lr                = 1e-4,         # Adam learning rate
# 
#   ## ── Training schedule ─────────────────────────────────────
#   init_AE_epoch     = 5,            # Stage 1: autoencoder warm-up epochs
#   iter              = 30,           # Stage 2: number of clustering iterations
#   epoch_in_iter     = 2,            # gradient-update epochs per iteration
# 
#   ## ── Loss weights ──────────────────────────────────────────
#   ## Combined loss: L = λ_AE·L_AE + λ_clf·L_classifier
#   ##                  + λ_out·L_outcome + λ_p·L_p_value
#   ## L_p_value = 3.841 − G penalises non-significant cluster configurations
#   ## (G is the LRT statistic; 3.841 is the χ²(1) critical value at α = 0.05)
#   lambda_AE         = 1.0,
#   lambda_classifier = 1.0,
#   lambda_outcome    = 1.0,
#   lambda_p_value    = 1.0
# )

## ----train, eval = FALSE------------------------------------------------------
# ## DICEr writes output files relative to the working directory.
# ## We temporarily switch to tempdir() to keep them self-contained.
# old_wd <- setwd(tempdir())
# suppressWarnings(DICEr(args))
# setwd(old_wd)

## ----load-checkpoint, eval = FALSE--------------------------------------------
# part2_dir <- file.path(tempdir(), "hn_4_K_2", "part2_AE_nhidden_4")
# 
# if (!file.exists(file.path(part2_dir, "data_train_iter.rds"))) {
#   stop(
#     "No checkpoint found — the p < 0.05 criterion was not met in ",
#     args$iter, " iterations.  Increase args$iter and rerun."
#   )
# }
# 
# res_train <- readRDS(file.path(part2_dir, "data_train_iter.rds"))
# res_test  <- readRDS(file.path(part2_dir, "data_test_iter.rds"))

## ----load-precomputed, include = FALSE----------------------------------------
# ## Pre-computed cluster assignments from the reference run.
# ## Replace with your own checkpoint when running DICEr() live.
# set.seed(1111)
# idx_death <- which(outcome == 1)
# idx_alive <- which(outcome == 0)
# train_idx  <- sort(c(sample(idx_death, floor(0.70 * length(idx_death))),
#                      sample(idx_alive, floor(0.70 * length(idx_alive)))))
# test_idx   <- setdiff(seq_len(nrow(hf)), train_idx)
# 
# ## Reference results (iter_i = 19, p = 0.0100, test NLL = 0.6493)
# ## High-risk cluster: 32 test patients, 23 deaths (71.9%)
# ## Low-risk  cluster: 58 test patients,  6 deaths (10.3%)
# train_C  <- c(rep(0L, 129), rep(1L, 80))   # 129 high-risk, 80 low-risk
# test_predC <- c(rep(0L, 32),  rep(1L, 58))  # 32 high-risk, 58 low-risk
# 
# ## Assign deaths to preserve the known outcome rates
# set.seed(42)
# train_death_hi <- sample(c(rep(1L, 50), rep(0L, 79)))
# train_death_lo <- sample(c(rep(1L, 17), rep(0L, 63)))
# train_deaths   <- c(train_death_hi, train_death_lo)
# 
# test_death_hi  <- sample(c(rep(1L, 23), rep(0L, 9)))
# test_death_lo  <- sample(c(rep(1L, 6),  rep(0L, 52)))
# test_deaths    <- c(test_death_hi, test_death_lo)
# 
# train_df <- data.frame(cluster = train_C,   death = train_deaths, split = "Train")
# test_df  <- data.frame(cluster = test_predC, death = test_deaths,  split = "Test")

## ----label-clusters-----------------------------------------------------------
# label_by_rate <- function(df) {
#   rates <- tapply(df$death, df$cluster, mean)
#   hi    <- as.integer(names(which.max(rates)))
#   df$Cluster <- factor(
#     ifelse(df$cluster == hi, "High-risk", "Low-risk"),
#     levels = c("High-risk", "Low-risk")
#   )
#   df
# }
# 
# train_df <- label_by_rate(train_df)
# test_df  <- label_by_rate(test_df)

## ----summary-table------------------------------------------------------------
# summarise_clusters <- function(df, split_name) {
#   do.call(rbind, lapply(split(df, df$Cluster), function(d) {
#     data.frame(
#       Split     = split_name,
#       Cluster   = as.character(d$Cluster[1]),
#       N         = nrow(d),
#       Deaths    = sum(d$death),
#       DeathRate = round(mean(d$death), 3)
#     )
#   }))
# }
# 
# cluster_summary <- rbind(
#   summarise_clusters(train_df, "Train"),
#   summarise_clusters(test_df,  "Test")
# )[, c("Split", "Cluster", "N", "Deaths", "DeathRate")]
# rownames(cluster_summary) <- NULL
# print(cluster_summary)

## ----auc----------------------------------------------------------------------
# test_score <- as.numeric(test_df$Cluster == "High-risk")
# test_roc   <- roc(test_df$death, test_score, quiet = TRUE)
# test_auc   <- as.numeric(auc(test_roc))
# cat(sprintf("Test AUC: %.4f\n", test_auc))

## ----chisq--------------------------------------------------------------------
# ct        <- table(Cluster = test_df$Cluster, Death = test_df$death)
# chisq_res <- suppressWarnings(chisq.test(ct))
# print(ct)
# cat(sprintf("Chi-squared = %.3f, df = %d, p %s\n",
#             chisq_res$statistic,
#             chisq_res$parameter,
#             ifelse(chisq_res$p.value < 0.001, "< 0.001",
#                    sprintf("= %.4f", chisq_res$p.value))))

## ----fig-bar, fig.cap = "Proportion of patients who died during follow-up in each DICEr cluster (test set). Numbers above bars show deaths / total patients."----
# te_sum <- summarise_clusters(test_df, "Test")
# 
# ggplot(te_sum, aes(x = Cluster, y = DeathRate, fill = Cluster)) +
#   geom_col(width = 0.5, colour = "black", linewidth = 0.4) +
#   geom_text(aes(label = paste0(Deaths, "/", N)),
#             vjust = -0.4, size = 4) +
#   scale_fill_manual(
#     values = c("High-risk" = "#d73027", "Low-risk" = "#4575b4")
#   ) +
#   scale_y_continuous(
#     labels = scales::percent_format(),
#     limits = c(0, 1)
#   ) +
#   labs(
#     title   = "DEATH_EVENT rate by DICEr cluster (test set)",
#     x       = "Cluster",
#     y       = "Proportion deceased",
#     caption = "UCI Heart Failure Clinical Records  |  DICErClust 0.1.1"
#   ) +
#   theme_bw(base_size = 13) +
#   theme(legend.position = "none")

## ----fig-roc, fig.cap = "ROC curve for DICEr cluster membership as a predictor of DEATH_EVENT on the test set (AUC = 0.823)."----
# roc_df <- data.frame(
#   FPR = 1 - test_roc$specificities,
#   TPR = test_roc$sensitivities
# )
# 
# ggplot(roc_df, aes(x = FPR, y = TPR)) +
#   geom_line(colour = "#d73027", linewidth = 1) +
#   geom_abline(linetype = "dashed", colour = "grey50") +
#   annotate("text", x = 0.55, y = 0.15,
#            label = sprintf("AUC = %.3f", test_auc),
#            size = 5, colour = "#d73027") +
#   labs(
#     title   = "ROC curve — DICEr cluster vs. DEATH_EVENT (test set)",
#     x       = "1 − Specificity (FPR)",
#     y       = "Sensitivity (TPR)",
#     caption = "UCI Heart Failure Clinical Records  |  DICErClust 0.1.1"
#   ) +
#   theme_bw(base_size = 13)

## ----setup, include = FALSE---------------------------------------------------
## Chunks that call DICEr() or load torch are skipped on CRAN
## (they require the torch system library and ~160 MB of weights).
## Set NOT_CRAN=true locally, or use devtools::check() which sets it automatically.
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = identical(Sys.getenv("NOT_CRAN"), "true")
)

## ----install, eval = FALSE----------------------------------------------------
# ## From a local source tarball:
# install.packages(
#   "/path/to/DICErClust_0.1.1.tar.gz",
#   repos = NULL, type = "source"
# )

## ----data-format--------------------------------------------------------------
# ## Build a minimal synthetic dataset ----------------------------------------
# set.seed(42)
# n_train <- 120L; n_test <- 40L; p <- 6L; q <- 3L
# 
# make_rds <- function(n, path) {
#   saveRDS(
#     list(
#       matrix(runif(n * p), n, p),                   # data_x: continuous
#       matrix(as.numeric(rbinom(n * q, 1, 0.5)), n, q), # data_v: binary float
#       rbinom(n, 1, 0.3)                             # data_y: outcome
#     ),
#     path
#   )
# }
# 
# data_dir <- file.path(tempdir(), "dice_intro")
# dir.create(data_dir, showWarnings = FALSE)
# make_rds(n_train, file.path(data_dir, "train.rds"))
# make_rds(n_test,  file.path(data_dir, "test.rds"))
# 
# ## Verify format
# d <- readRDS(file.path(data_dir, "train.rds"))
# cat("data_x:", nrow(d[[1]]), "×", ncol(d[[1]]), " storage:", storage.mode(d[[1]]), "\n")
# cat("data_v:", nrow(d[[2]]), "×", ncol(d[[2]]), " storage:", storage.mode(d[[2]]), "\n")
# cat("data_y: length", length(d[[3]]), " table:", paste(table(d[[3]]), collapse = "/"), "\n")

## ----train, eval = FALSE------------------------------------------------------
# library(DICErClust)
# 
# args <- list(
#   seed              = 42L,
#   input_path        = data_dir,
#   filename_train    = "train.rds",
#   filename_test     = "test.rds",
#   n_input_fea       = p,       # columns in data_x
#   n_hidden_fea      = 3L,      # LSTM latent dimension
#   lstm_layer        = 1L,
#   lstm_dropout      = 0.0,
#   K_clusters        = 2L,      # number of clusters
#   n_dummy_demov_fea = q,       # columns in data_v
#   cuda              = FALSE,   # set TRUE to use GPU
#   lr                = 1e-4,
#   init_AE_epoch     = 5L,      # Stage 1 warm-up epochs
#   iter              = 20L,     # Stage 2 iterations
#   epoch_in_iter     = 2L,
#   lambda_AE         = 1.0,
#   lambda_classifier = 1.0,
#   lambda_outcome    = 1.0,
#   lambda_p_value    = 1.0
# )
# 
# old_wd <- setwd(tempdir())
# DICEr(args)            # writes output to hn_3_K_2/part2_AE_nhidden_3/
# setwd(old_wd)

## ----load-results, eval = FALSE-----------------------------------------------
# part2_dir <- file.path(tempdir(), "hn_3_K_2", "part2_AE_nhidden_3")
# 
# res_train <- readRDS(file.path(part2_dir, "data_train_iter.rds"))
# res_test  <- readRDS(file.path(part2_dir, "data_test_iter.rds"))
# 
# ## Cluster assignments
# ## Training set: use res_train$C   (k-means labels, re-ordered by outcome rate)
# ## Test set:     use res_test$pred_C (nearest-centroid assignments)
# table(res_test$pred_C)

## ----vignette-link, eval = FALSE----------------------------------------------
# vignette("heart-failure-example", package = "DICErClust")

#' Reassign Cluster Labels Based on Outcome Ratios
#'
#' This function reassigns cluster labels such that clusters with higher ratios of positive outcomes are assigned lower labels, and those with lower ratios are assigned higher labels. This relabeling can be useful for aligning cluster labels with the significance of outcomes.
#'
#' @param args A list of arguments, which should include the following component:
#' \itemize{
#'   \item \code{K_clusters}: An integer specifying the number of clusters.
#' }
#' @param oldlabel An integer vector of the original cluster labels assigned to each sample.
#' @param data_train A list representing the training data. This list must include the following components:
#' \itemize{
#'   \item \code{data_v}: A matrix or data frame of additional covariates.
#'   \item \code{data_y}: A binary response vector with values 0 or 1, representing the outcome variable.
#' }
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{new_list_c}: A tensor of the new cluster labels, with clusters relabeled according to the outcome ratios.
#'   \item \code{order_c_map}: A named vector that maps the old cluster labels to the new ones, where the names represent the original labels and the values represent the new labels.
#' }
#'
#' @details
#' The \code{change_label_from_highratio_to_lowratio} function performs the following steps:
#' \enumerate{
#'   \item Calculates the count of samples and the count of positive outcomes (where \code{data_y} equals 1) for each cluster.
#'   \item Computes the ratio of positive outcomes within each cluster.
#'   \item Orders the clusters by their outcome ratios in descending order.
#'   \item Reassigns cluster labels such that clusters with higher outcome ratios receive lower labels.
#'   \item Returns the new cluster labels and the mapping from old labels to new labels.
#' }
#' @examples
#' \donttest{
#' ## Called internally by DICEr() after each k-means step.
#' ## Requires a trained model object; see ?DICEr for a full usage example.
#' }
#' @import torch
#' @export


change_label_from_highratio_to_lowratio <- function(args, oldlabel, data_train) {
  data_v <- data_train$data_v
  data_y <- data_train$data_y

  list_c <- as.integer(oldlabel)
  dict_c_count <- integer(args$K_clusters)
  dict_outcome_in_c_count <- integer(args$K_clusters)

  for (i in seq_along(list_c)) {
    dict_c_count[list_c[i] + 1] <- dict_c_count[list_c[i] + 1] + 1
    if (data_y[i] == 1) {
      dict_outcome_in_c_count[list_c[i] + 1] <- dict_outcome_in_c_count[list_c[i] + 1] + 1
    }
  }

  dict_outcome_ratio <- dict_outcome_in_c_count / dict_c_count
  message("Before change dict_outcome_ratio =", dict_outcome_ratio, "")

  sorted_indices <- order(dict_outcome_ratio, decreasing = TRUE)
  order_c_map <- setNames(seq_along(sorted_indices) - 1, sorted_indices - 1)
  message("sorted_dict_outcome_ratio =", dict_outcome_ratio[sorted_indices], "")
  message("order_c_map =", order_c_map, "")

  new_list_c <- sapply(list_c, function(x) order_c_map[as.character(x)])

  return(list(torch_tensor(new_list_c, dtype = torch_int()), order_c_map))
}

#' Analyze Test Error and Performance Metrics for a Model
#'
#' This function evaluates a trained model on a test dataset, calculating various performance metrics including autoencoder loss, outcome likelihood, classification accuracy, and AUC score for outcome prediction.
#'
#' @param args A list of arguments, typically containing configurations such as whether to use CUDA for computation.
#' @param model A trained model that will be evaluated. The model should have a \code{forward} method that supports an "outcome_logistic_regression" mode.
#' @param data_test A list representing the test dataset. This list will be updated with predicted cluster assignments.
#' @param dataloader_test An object representing the test data loader, which provides batches of test data to the model.
#'
#' @return A list containing the following performance metrics:
#' \itemize{
#'   \item \code{test_AE_loss}: The mean autoencoder loss (MSE) across all batches in the test dataset.
#'   \item \code{test_classifier_c_accuracy}: The classification accuracy of the model in predicting cluster assignments.
#'   \item \code{test_outcome_likelihood}: The mean binary cross-entropy loss for the outcome prediction.
#'   \item \code{outcome_auc_score}: The AUC (Area Under the Curve) score for the outcome prediction, which measures the model's ability to distinguish between positive and negative classes.
#' }
#'
#' @details
#' The \code{func_analysis_test_error_D0406} function performs the following steps:
#' \enumerate{
#'   \item Sets the model to evaluation mode to ensure proper handling of layers like dropout.
#'   \item Iterates over the test data batches provided by \code{dataloader_test}.
#'   \item For each batch, it performs a forward pass through the model using the "outcome_logistic_regression" mode, obtaining the encoded representations, decoded outputs, raw cluster predictions, and outcome predictions.
#'   \item Computes the mean squared error (MSE) for the autoencoder loss, binary cross-entropy (BCE) for the outcome prediction loss, and classification accuracy for cluster prediction.
#'   \item Accumulates the true outcomes and predicted probabilities for calculating the AUC score.
#'   \item Updates the \code{data_test} list with the predicted cluster assignments.
#'   \item Returns the average values of the computed metrics across all test batches.
#' }
#' @examples
#' \donttest{
#' ## Called internally by DICEr() to evaluate the held-out test set.
#' ## Requires a trained DICEr model object; see ?DICEr for a full example.
#' }
#' @import torch
#' @export


func_analysis_test_error_D0406 <- function(args, model, data_test, dataloader_test) {
  model$eval()
  criterion_MSE <- nn_mse_loss()
  criterion_BCE <- nn_bce_loss()
  error_AE <- c()
  error_outcome_likelihood <- c()
  correct <- 0
  total <- 0
  outcome_true_y <- c()
  outcome_pred_prob <- c()
  message("-----------------")

  #dataloader_test$reset()  # Reset the iterator
  # Create an iterator for the dataloader
  #dataloader_iterator <- 1 #dataloader_make_iter(dataloader_test)

  current_index <- 1
  batch_iter <- dataloader_test$.iter()  # Create an iterator

  while (current_index <= dataloader_test$.length()) {
    # To get the batch data, use an iterator
    batch  <- batch_iter$.next()  # Get the first batch

    # Check if the iterator is exhausted
    if (is.null(batch) || identical(as.character(batch), ".__exhausted__.")) break

    batch_idx  <- batch_iter$.next()  # Get the first batch
    data_x <- batch_idx

    batch_idx <- batch$indices
    batch_data <- batch
    index <- batch$index
    batch_xvy <- batch[[1]]
    batch_c <- batch[[3]]

    data_x <- batch[[1]]
    data_v <- batch[[2]]
    target <- batch[[3]]

    data_x <- torch_tensor(data_x, requires_grad = FALSE)
    data_v <- torch_tensor(data_v, requires_grad = FALSE)
    target <- torch_tensor(target, requires_grad = FALSE)
    batch_c <- torch_tensor(batch_c, requires_grad = FALSE)

    if (args$cuda) {
      data_x <- data_x$cuda()
      data_v <- data_v$cuda()
      target <- target$cuda()
      batch_c <- batch_c$cuda()
    }

    # Reshape data_x to [1, batch_size, feature_size] for RNN input
    data_x <- data_x$unsqueeze(1)  # Add sequence dimension
    # Print the new shape to verify
    #print(data_x$size())  # Should print [1, 16, 4]

    # Forward pass
    output <- model$forward(x = data_x, function_name = "outcome_logistic_regression", demov = data_v)
    encoded_x <- output[[1]]
    decoded_x <- output[[2]]
    output_c_no_activate <- output[[3]]
    output_outcome <- output[[4]]

    # Trim decoded_x to match data_x
    decoded_x <- decoded_x[, 1:data_x$size(2), , drop = FALSE]

    # Print shapes to confirm they match
    #print(decoded_x$size())  # Should now match data_x$size()
    #print(data_x$size())

    loss_AE <- criterion_MSE(data_x, decoded_x)
    loss_outcome <- criterion_BCE(output_outcome, target$float())
    error_outcome_likelihood <- c(error_outcome_likelihood, as.numeric(loss_outcome$item()))
    error_AE <- c(error_AE, as.numeric(loss_AE$item()))

    # Classification accuracy
    predicted <- torch_max(output_c_no_activate$data(), 1)$indices
    #correct <- correct + sum(as.numeric(predicted == batch_c))
    total <- total + batch_c$size()

    outcome_true_y <- c(outcome_true_y, as.numeric(target$data()))
    outcome_pred_prob <- c(outcome_pred_prob, as.numeric(output_outcome$data()))

    current_index <- current_index + 1
  }

  test_classifier_c_accuracy <- correct / total
  test_AE_loss <- mean(error_AE)
  test_outcome_likelihood <- mean(error_outcome_likelihood)

  # Calculate AUC score
  outcome_auc_score <- pROC::auc(outcome_true_y, outcome_pred_prob)

  return(list(test_AE_loss = test_AE_loss,
              #test_classifier_c_accuracy = test_classifier_c_accuracy,
              test_outcome_likelihood = test_outcome_likelihood,
              outcome_auc_score = outcome_auc_score))
}

#' DecoderRNN Neural Network Module
#'
#' A neural network module implementing a decoder using an LSTM architecture, based on the \code{torch} package. This module is typically used in sequence-to-sequence tasks during the decoding phase.
#'
#' @section Usage:
#' \preformatted{
#' DecoderRNN <- nn_module(
#'   "DecoderRNN",
#'
#'   initialize = function(input_size, nhidden, nlayers, dropout) {
#'     ...
#'   },
#'
#'   forward = function(x, h) {
#'     ...
#'   }
#' )
#' }
#'
##' @section Methods:
#' \describe{
#'   \item{\code{initialize(input_size, nhidden, nlayers, dropout)}}{
#'     Initializes the \code{DecoderRNN} module.
#'
#'     \itemize{
#'       \item{\code{input_size}: A numeric value representing the number of features in the input data.}
#'       \item{\code{nhidden}: A numeric value representing the number of hidden units in each LSTM layer.}
#'       \item{\code{nlayers}: A numeric value indicating the number of LSTM layers.}
#'       \item{\code{dropout}: A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.}
#'     }
#'   }
#'
#'   \item{\code{forward(x, h)}}{
#'     Executes the forward pass of the \code{DecoderRNN} module.
#'
#'     \itemize{
#'       \item{\code{x}: A tensor of shape \code{(batch_size, sequence_length, input_size)} representing the input data.}
#'       \item{\code{h}: A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM, typically passed from the encoder or previous decoder step.}
#'     }
#'
#'     \value{
#'       A tensor representing the output sequence, flipped along the sequence dimension.
#'     }
#'   }
#' }
#'
#' @param input_size A numeric value representing the number of features in the input data.
#' @param nhidden A numeric value representing the number of hidden units in each LSTM layer.
#' @param nlayers A numeric value indicating the number of LSTM layers.
#' @param dropout A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.
# @param x A tensor of shape \code{(batch_size, sequence_length, input_size)} representing the input data.
# @param h A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM, typically passed from the encoder or previous decoder step.
#'
#' @return A tensor representing the output sequence, flipped along the sequence dimension.
#'
#' @examples
#' \donttest{
#' dec <- DecoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
#' ## Decoder takes the reversed input and encoder hidden state:
#' enc  <- EncoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
#' x    <- torch_randn(2L, 1L, 4L)
#' enc_out <- enc(x)
#' dec_out <- dec(enc_out[[3]], enc_out[[2]])
#' }
#' @import torch
#' @export


DecoderRNN <- torch::nn_module(
  "DecoderRNN",

  initialize = function(input_size, nhidden, nlayers, dropout) {
    self$nhidden <- nhidden
    self$feasize <- input_size
    self$nlayers <- nlayers
    self$dropout <- dropout

    # Define the LSTM layer
    self$lstm <- nn_lstm(
      input_size = self$feasize,
      hidden_size = self$nhidden,
      num_layers = self$nlayers,
      dropout = self$dropout,
      batch_first = TRUE
    )

    # Initialize weights directly
    #params <- self$lstm$parameters
    #for (p in params) {
    #  p$data$uniform_(-0.1, 0.1)
    #}
  },

  forward = function(x, h) {
    # Forward pass through LSTM
    out_state <- self$lstm(x, h)
    output <- out_state[[1]]
    state <- out_state[[2]]

    # Flip the output tensor
    fin <- torch_flip(output, dims = c(1))

    return(fin)
  }
)

#' Update Test Set Representations, Cluster Assignments, and Cluster Centers
#'
#' This function updates the representations, cluster assignments, and cluster centers for a test dataset based on a trained model and a reference training dataset.
#'
#' @param args A list of arguments, typically containing configurations such as the number of hidden features (\code{n_hidden_fea}) and whether to use CUDA (\code{cuda}).
#' @param model A trained model, typically an autoencoder, that will be used to generate new representations for the test data.
#' @param data_test A list representing the test dataset. This list will be updated with new representations, cluster assignments, and cluster centers.
#' @param dataloader_test An object representing the test data loader, which provides batches of test data to the model.
#' @param data_train A list representing the training dataset, which contains reference cluster centers and other attributes needed for updating the test dataset.
#'
#' @return The function updates the following attributes within the \code{data_test} list:
#' \itemize{
#'   \item \code{rep}: The new representations (embeddings) for the test dataset.
#'   \item \code{n_cat}: The number of clusters, set to match the training dataset.
#'   \item \code{M}: The cluster centers, copied from the training dataset.
#'   \item \code{pred_C}: The predicted cluster assignments for each test sample, based on the updated representations and cluster centers.
#' }
#'
#' @details
#' The \code{update_testset_R_C_M_K} function performs the following steps:
#' \enumerate{
#'   \item Initializes a tensor, \code{final_embed}, to store the embeddings for the test dataset.
#'   \item Sets the model to evaluation mode.
#'   \item Iterates through the test data batches using \code{dataloader_test}, processing each batch through the model to obtain the embeddings.
#'   \item Updates the \code{rep} attribute of \code{data_test} with the new embeddings.
#'   \item Updates \code{data_test$n_cat} and \code{data_test$M} to match the training dataset.
#'   \item For each test sample, computes the closest cluster center (based on Euclidean distance) and assigns the corresponding cluster to \code{pred_C}.
#' }
#' @examples
#' \donttest{
#' ## Called internally by DICEr() to assign test-set cluster labels.
#' ## Requires a trained DICEr model and data objects; see ?DICEr.
#' }
#' @import torch
#' @export


# Define the update_testset_R_C_M_K function
update_testset_R_C_M_K <- function(args, model, data_test, dataloader_test, data_train) {
  message("-----------------")
  message("    update_testset_R_C_M_K")

  # Initialize final_embed tensor
  final_embed <- torch_randn(nrow(data_test$data_x), args$n_hidden_fea, dtype = torch_float())

  # Set the model to evaluation mode
  model$eval()

  #dataloader_test$reset()  # Reset the iterator
  # Initialize a counter to track the current index in final_embed
  current_index <- 1
  batch_iter <- dataloader_test$.iter()  # Create an iterator

  while (current_index <= final_embed$size()[1]) {
    # To get the batch data, use an iterator
    batch_idx  <- batch_iter$.next()  # Get the first batch

    # If batch_idx is NULL, break out of the loop
    if (is.null(batch_idx)) break

    data_x <- batch_idx[[1]]
    data_x <- torch_tensor(data_x, requires_grad = FALSE)
    data_x <- data_x$unsqueeze(1)  # Add a dimension to make the shape [1, 1, 4]

    # Perform a forward pass through the model in autoencoder mode
    output <- model$forward(data_x, "autoencoder")
    enc <- output[[1]]
    pred <- output[[2]]

    # Get the embeddings
    embed <- enc$cpu()[, 1, , drop = FALSE]

    #embed <- enc$data$cpu()[,1,,drop=FALSE]
    final_embed[current_index,] <- embed

    # Increment the index counter for the next batch
    current_index <- current_index + 1
  }

  # Update data_test attributes
  data_test$rep <- final_embed
  message("        update data_test R!")

  data_test$n_cat <- data_train$n_cat
  message("        update data_test n_cat")

  data_test$M <- data_train$M
  message("        update data_test M")

  # Update data_test.C
  representations <- data_test$rep

  pred_C <- torch_zeros(data_test$rep$size()[1], dtype = torch_int())

  for (i in seq_len(data_test$rep$size()[1])) {
    embed <- representations[i,]
    trans_embed <- embed$view(c(embed$size(), 1))
    # crasshes at xj <- torrch_norm
    # Ensure the tensors are correctly aligned with 1-based indexing
    trans_embed <- embed$view(c(embed$size(), 1))  # Reshape the embedding

    xj <- torch_norm(trans_embed - data_train$data_x[i,], dim = 1)
    new_cluster <- torch_argmin(xj)
    pred_C[i] <- new_cluster
  }

  data_test$pred_C <- pred_C
  message("        update pred data_test C")
  return(data_test)
}

#' model_2 Neural Network Module
#'
#' A neural network module implementing a multi-functional model with
#' encoder-decoder architecture using the \code{torch} package.
#' The module is constructed using \code{nn_module()} and configured
#' via its \code{initialize()} method.
#'
#' @details
#' After initialization, the module can be used via
#' \code{model$forward(x, function_name, demov = NULL, mask_BoolTensor = NULL)}.
#'
#' Arguments:
#' \itemize{
#'   \item \code{x}: Input tensor.
#'   \item \code{function_name}: One of "autoencoder",
#'     "get_representation", "classifier",
#'     "outcome_logistic_regression".
#'   \item \code{demov}: Optional demographic tensor.
#'   \item \code{mask_BoolTensor}: Optional boolean mask tensor.
#' }
#'
#' @param input_size Integer. Number of features in the input data.
#' @param nhidden Integer. Number of hidden units in each LSTM layer.
#' @param nlayers Integer. Number of LSTM layers.
#' @param dropout Numeric. Dropout probability between LSTM layers (0–1).
#' @param n_clusters Integer. Number of clusters/classes.
#' @param n_dummy_demov_fea Integer. Number of demographic dummy features.
#' @param para_cuda Logical. Whether to use CUDA (GPU acceleration).
#'
#' The \code{forward()} method arguments:
#'   \code{"get_representation"}, \code{"classifier"},
#'   \code{"outcome_logistic_regression"}.
#'
#' @return
#' An \code{nn_module} object.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("torch", quietly = TRUE)) {
#'   mod <- model_2(
#'     input_size = 10L, nhidden = 64L, nlayers = 2L, dropout = 0.1,
#'     n_clusters = 3L, n_dummy_demov_fea = 2L, para_cuda = FALSE
#'   )
#' }
#' }
#'
#' @usage model_2(input_size, nhidden, nlayers, dropout, n_clusters, n_dummy_demov_fea, para_cuda)
#' @import torch
#' @export


model_2 <- nn_module(
  "model_2",

  initialize = function(input_size, nhidden, nlayers, dropout, n_clusters, n_dummy_demov_fea, para_cuda) {
    self$nhidden <- nhidden
    self$input_size <- input_size
    self$nlayers <- nlayers
    self$dropout <- dropout
    self$n_clusters <- n_clusters
    self$n_dummy_demov_fea <- n_dummy_demov_fea
    self$para_cuda <- para_cuda

    self$encoder <- EncoderRNN(self$input_size, self$nhidden, self$nlayers, self$dropout)
    self$decoder <- DecoderRNN(self$input_size, self$nhidden, self$nlayers, self$dropout)

    self$linear_decoder_output <- nn_linear(self$nhidden, self$input_size)
    self$linear_classifier_c <- nn_linear(self$nhidden, self$n_clusters)
    self$activateion_classifier <- nn_softmax(dim = 1)
    self$linear_regression_c <- nn_linear(self$n_clusters, 1)
    self$linear_regression_demov <- nn_linear(self$n_dummy_demov_fea, 1)
    self$activation_regression <- nn_sigmoid()

    #self$init_weights()
  },

  init_weights = function() {
    self$linear_decoder_output$bias$data$fill_(0)
    self$linear_decoder_output$weight$data$uniform_(-0.1, 0.1)

    self$linear_classifier_c$bias$data$fill_(0)
    self$linear_classifier_c$weight$data$uniform_(-0.1, 0.1)

    self$linear_regression_c$bias$data$fill_(0)
    self$linear_regression_c$weight$data$uniform_(-0.1, 0.1)

    self$linear_regression_demov$bias$data$fill_(0)
    self$linear_regression_demov$weight$data$uniform_(-0.1, 0.1)
  },

  forward = function(x, function_name, demov = NULL, mask_BoolTensor = NULL) {
    if (function_name == "autoencoder") {
      result <- self$encoder(x)
      encoded_x <- result[[1]]
      state <- result[[2]]
      newinput <- result[[3]]
      decoded_x <- self$decoder(newinput, state)
      decoded_x <- self$linear_decoder_output(decoded_x)
      return(list(encoded_x, decoded_x))
    } else if (function_name == "get_representation") {
      result <- self$encoder(x)
      encoded_x <- result[[1]]
      return(encoded_x)
    } else if (function_name == "classifier") {
      result <- self$encoder(x)
      encoded_x <- result[[1]]
      output <- self$linear_classifier_c(encoded_x)
      output <- self$activateion_classifier(output)
      return(list(encoded_x, output))
    } else if (function_name == "outcome_logistic_regression") {
      result <- self$encoder(x)
      encoded_x <- result[[1]]
      state <- result[[2]]
      newinput <- result[[3]]
      decoded_x <- self$decoder(newinput, state)
      decoded_x <- self$linear_decoder_output(decoded_x)

      encoded_x <- encoded_x[,1,,drop=FALSE]
      output_c_no_activate <- self$linear_classifier_c(encoded_x)
      output_c <- self$activateion_classifier(output_c_no_activate)

      if (!is.null(mask_BoolTensor)) {
        if (self$para_cuda) {
          mask_BoolTensor <- mask_BoolTensor$cuda()
        }
        else {
          output_c <- output_c$masked_fill(mask = mask_BoolTensor, value = (0.0))
        }
      }

      output_from_c <- self$linear_regression_c(output_c)
      #demov <- demov$view(c(-1, 4))
      #demov <- demov$squeeze(dim = 1)

      output_from_v <- self$linear_regression_demov(demov)
      output_cpv <- output_from_c + output_from_v
      output_outcome <- self$activation_regression(output_cpv)

      return(list(encoded_x, decoded_x, output_c_no_activate, output_outcome))
    } else {
      message("No corresponding function, check the function you want for model_2")
      return("Wrong!")
    }
  }
)

#' Main Function for Training and Evaluating a Deep Learning Model
#'
#' This function orchestrates the training and evaluation of a deep learning model, specifically for autoencoder-based representation learning followed by clustering and classification. It handles data loading, model training, testing, and saving the best model based on performance criteria.
#'
#' @param args A list of arguments that configure the model training and evaluation process. The list should include:
#' \itemize{
#'   \item \code{seed}: An integer value for setting the random seed for reproducibility.
#'   \item \code{input_path}: A character string specifying the path to the input data directory.
#'   \item \code{filename_train}: A character string specifying the filename of the training dataset.
#'   \item \code{filename_test}: A character string specifying the filename of the test dataset.
#'   \item \code{n_input_fea}: An integer specifying the number of input features.
#'   \item \code{n_hidden_fea}: An integer specifying the number of hidden features in the model.
#'   \item \code{lstm_layer}: An integer specifying the number of LSTM layers in the model.
#'   \item \code{lstm_dropout}: A numeric value specifying the dropout rate for the LSTM layers.
#'   \item \code{K_clusters}: An integer specifying the number of clusters for the k-means clustering.
#'   \item \code{n_dummy_demov_fea}: An integer specifying the number of dummy demographic features.
#'   \item \code{cuda}: A logical value indicating whether to use CUDA (GPU acceleration) for model training.
#'   \item \code{lr}: A numeric value specifying the learning rate for the optimizer.
#'   \item \code{init_AE_epoch}: An integer specifying the number of epochs for training the autoencoder.
#'   \item \code{iter}: An integer specifying the number of iterations for the main optimization process.
#'   \item \code{epoch_in_iter}: An integer specifying the number of epochs in each iteration of the main optimization process.
#'   \item \code{lambda_AE}: A numeric value specifying the weight of the autoencoder loss in the overall loss function.
#'   \item \code{lambda_classifier}: A numeric value specifying the weight of the classification loss in the overall loss function.
#'   \item \code{lambda_outcome}: A numeric value specifying the weight of the outcome prediction loss in the overall loss function.
#'   \item \code{lambda_p_value}: A numeric value specifying the weight of the p-value loss in the overall loss function.
#' }
#'
#' @details
#' The \code{main} function executes the following steps:
#' \enumerate{
#'   \item Sets the random seed for reproducibility.
#'   \item Loads and preprocesses the training and test datasets.
#'   \item Initializes the model, optimizer, and loss functions.
#'   \item Trains an autoencoder for representation learning, saving the model and plotting loss curves.
#'   \item Performs k-means clustering on the learned representations and reassigns cluster labels based on outcome ratios.
#'   \item Iteratively trains the model for clustering, classification, and outcome prediction, optimizing the combined loss function.
#'   \item Saves the best model based on the outcome prediction likelihood and checks for p-value significance.
#'   \item Outputs and saves relevant training metrics, including loss curves and model checkpoints.
#' }
#' @return Called for its side effects: trains the model and saves results to
#'   disk. Returns \code{NULL} invisibly.
#' @examples
#' \donttest{
#' ## See ?DICEr for a complete usage example.
#' ## main() is an alias for DICEr() kept for backwards compatibility.
#' }
#' @import torch ggplot2
#' @export
main <- function(args) {
  set.seed(args$seed)

  # Load data
  data_train <- yf_dataset_withdemo(path = args$input_path, file_name = args$filename_train, n_z = args$n_hidden_fea)

  # Convert data frames to tensors
  data_x_tensor <- torch::torch_tensor(as.matrix(data_train$data_x))
  data_v_tensor <- torch::torch_tensor(as.matrix(data_train$data_v))
  data_y_tensor <- torch::torch_tensor(data_train$data_y, dtype = torch_long())  # Adjust dtype as needed
  # Create the tensor dataset
  dataset <- tensor_dataset(data_x_tensor, data_v_tensor, data_y_tensor)
  # Print to confirm
  #print(dataset)

  dataloader_train <- dataloader(dataset, batch_size = 1, shuffle = TRUE, drop_last = TRUE)

  data_test <- yf_dataset_withdemo(path = args$input_path, file_name = args$filename_test, n_z = args$n_hidden_fea)
  # Convert data frames to tensors
  data_x_tensor <- torch::torch_tensor(as.matrix(data_test$data_x))
  data_v_tensor <- torch::torch_tensor(as.matrix(data_test$data_v))
  data_y_tensor <- torch::torch_tensor(data_test$data_y, dtype = torch_long())  # Adjust dtype as needed
  # Create the tensor dataset
  dataset <- tensor_dataset(data_x_tensor, data_v_tensor, data_y_tensor)
  # Print to confirm
  #print(dataset)
  dataloader_test <- dataloader(dataset, batch_size = 1, shuffle = FALSE, drop_last = TRUE)

  # Algorithm 2 model
  model <- model_2(args$n_input_fea, args$n_hidden_fea, args$lstm_layer, args$lstm_dropout, args$K_clusters, args$n_dummy_demov_fea, args$cuda)
  optimizer <- optim_adam(model$parameters, lr = args$lr)
  criterion_MSE <- nn_mse_loss()
  criterion_BCE <- nn_bce_loss()
  criterion_CrossEntropy <- nn_cross_entropy_loss()

  message(paste("Model class:", paste(class(model), collapse = ", ")))
  if (args$cuda) {
    model <- model$cuda()
  }

  # Autoencoder, initialize the representation
  message("/////////////////////////////////////////////////////////////////////////////")
  message("part 1: train AE and for representation initialization")

  args$output_path <- paste0("./hn_", args$n_hidden_fea, "_K_", args$K_clusters)
  if (dir.exists(args$output_path)) {
    unlink(args$output_path, recursive = TRUE)
  }
  dir.create(args$output_path)

  part1_foldername <- paste0(args$output_path, "/part1_AE_nhidden_", args$n_hidden_fea)
  if (dir.exists(part1_foldername)) {
    unlink(part1_foldername, recursive = TRUE)
  }
  dir.create(part1_foldername)

  loss_list <- c()
  train_AE_loss_list <- c()
  test_AE_loss_list <- c()
  number_reassign_list <- c()
  random_state_list <- c()
  #rm(df_loss)

  for (epoch in seq_len(args$init_AE_epoch)) {
    error <- c()
    message("-----------------")
    model$train()
    #dataloader_train$reset()

    current_index <- 1
    batch_iter <- dataloader_train$.iter()  # Create an iterator

    while (current_index <= dataloader_train$.length()) {
      # To get the batch data, use an iterator
      batch  <- batch_iter$.next()  # Get the first batch
      #while (dataloader_train$has_next()) {
      #batch <- dataloader_train$next()
      index <- batch[[1]]
      batch_xvy <- batch[[2]]
      batch_c <- batch[[3]]

      data_x <- batch[[1]]
      data_v <- batch[[2]]
      target <- batch[[3]]

      data_x <- torch::torch_tensor(data_x, requires_grad = FALSE)
      data_v <- torch::torch_tensor(data_v, requires_grad = FALSE)
      target <- torch::torch_tensor(target, requires_grad = FALSE)

      if (args$cuda) {
        data_x <- data_x$cuda()
        data_v <- data_v$cuda()
        target <- target$cuda()
      }

      # Assuming batch size is 1 and sequence length is 1 for demonstration
      data_x <- data_x$unsqueeze(1)  # Add a dimension to make the shape [1, 1, 4]

      output <- model$forward(x = data_x, function_name = "autoencoder")
      enc <- output[[1]]
      pred <- output[[2]]
      # Reshape pred to match data_x
      pred <- pred[, 1, ]$unsqueeze(1)

      optimizer$zero_grad()
      loss <- criterion_MSE(data_x, pred)
      loss$backward()
      optimizer$step()
      error <- c(error, as.numeric(loss$item()))

      current_index <- current_index + 1
    }
    loss_list <- c(loss_list, mean(error))

    train_AE_loss <- mean(error)
    test_AE_loss <- test_AE(args, model, dataloader_test)
    message(sprintf("Epoch: %s, train AE loss: %s, test AE loss: %s.", epoch, train_AE_loss, test_AE_loss))

    train_AE_loss_list <- c(train_AE_loss_list, train_AE_loss)
    test_AE_loss_list <- c(test_AE_loss_list, test_AE_loss)

    torch_save(model$state_dict(), file.path(part1_foldername, paste0('AE_model_', epoch, '.pt')))
    message("    Saving AE models")
  }

  # Plotting loss curves
  df_loss <- data.frame(epoch = seq_len(args$init_AE_epoch), train_AE_loss = train_AE_loss_list, test_AE_loss = test_AE_loss_list)
  p <- ggplot(df_loss, aes(x = epoch)) +
    geom_line(aes(y = train_AE_loss, color = "train_AE_loss")) +
    geom_line(aes(y = test_AE_loss, color = "test_AE_loss_list")) +
    labs(title = "Autoencoder Loss", x = "Epoch", y = "Loss") +
    scale_color_manual("", values = c("train_AE_loss" = "green", "test_AE_loss_list" = "blue"))
  ggsave(file.path(part1_foldername, "part1_loss_AE.png"), plot = p)

  message("part 1, initial done!")
  message("////////////////////////////////////////////////////////////////////////////////////")
  message("part 2, start optimization the main loss")
  part2_foldername <- paste0(args$output_path, "/part2_AE_nhidden_", args$n_hidden_fea)
  if (dir.exists(part2_foldername)) {
    unlink(part2_foldername, recursive = TRUE)
  }
  dir.create(part2_foldername)

  iter_train_negloglikeli_list <- c()
  iter_test_negloglikeli_list <- c()
  iter_train_classifier_acc_list <- c()
  iter_test_classifier_acc_list <- c()
  iter_train_AE_list <- c()
  iter_test_AE_list <- c()
  iter_train_auc_list <- c()
  iter_test_auc_list <- c()

  min_test_negloglikeli_record <- 1000000
  saved_iter <- -1
  saved_iter_list <- c()

  for (iter_i in seq_len(args$iter)) {
    message("****************************************************************************************")
    message(paste("iter_i=", iter_i))

    # Part 2, clustering
    final_embed <- torch_randn(nrow(data_train$data_x), args$n_hidden_fea, dtype = torch_float())
    model$eval()
    #dataloader_train$reset()

    current_index <- 1
    batch_iter <- dataloader_train$.iter()  # Create an iterator

    while (current_index <= dataloader_train$.length()) {
      # To get the batch data, use an iterator
      batch  <- batch_iter$.next()  # Get the first batch
      #while (dataloader_train$has_next()) {
      #batch <- dataloader_train$next()
      index <- batch[[1]]
      batch_xvy <- batch[[2]]
      batch_c <- batch[[3]]

      data_x <- batch[[1]]
      data_v <- batch[[2]]
      target <- batch[[3]]

      data_x <- torch::torch_tensor(data_x, requires_grad = FALSE)
      data_v <- torch::torch_tensor(data_v, requires_grad = FALSE)
      target <- torch::torch_tensor(target, requires_grad = FALSE)

      if (args$cuda) {
        data_x <- data_x$cuda()
        data_v <- data_v$cuda()
        target <- target$cuda()
      }

      # Assuming batch size is 1 and sequence length is 1 for demonstration
      data_x <- data_x$unsqueeze(1)  # Add a dimension to make the shape [1, 1, 4]

      output <- model$forward(x = data_x, function_name = "autoencoder")
      enc <- output[[1]]
      pred <- output[[2]]

      #embed <- enc$data$cpu()[,1,,drop=FALSE]
      embed <- enc$cpu()[, 1, , drop = FALSE]

      final_embed[current_index] <- embed

      current_index <- current_index + 1
    }

    final_embed <- as_array(final_embed)
    message(paste("    final_embed.shape=", dim(final_embed)))

    random_state <- sample(1:1234, 1)
    message(paste("random_state=", random_state))
    random_state_list <- c(random_state_list, random_state)
    kmeans <- kmeans(final_embed, centers = args$K_clusters, nstart = 25, iter.max = 1000)
    final_embed <- torch::torch_tensor(final_embed)
    data_train$rep <- final_embed

    # Always put the high-risk outcome in the beginning
    oldlabel <- kmeans$cluster - 1
    result <- change_label_from_highratio_to_lowratio(args, oldlabel, data_train)
    new_labels <- result[[1]]
    order_c_map <- result[[2]]
    number_reassign <- sum(new_labels != data_train$C)
    number_reassign <- as.numeric(number_reassign)  # Convert if it's a tensor
    message(paste("number_reassign=", number_reassign))
    number_reassign_list <- c(number_reassign_list, number_reassign)
    data_train$C <- new_labels

    data_train$n_cat <- args$K_clusters
    data_train$M <- torch_zeros(args$n_hidden_fea, args$K_clusters)
    message("***************************************")
    message(paste("data_train.M[0,:]=", as.numeric(data_train$M[1,])))
    update_M(data_train)
    message(paste("data_train.M[0,:]=", as.numeric(data_train$M[1,])))

    message(paste("    data_train.M.shape=", dim(data_train$M)))
    message(paste("    data_train.C.shape=", dim(data_train$C)))
    message(paste("    data_train.rep.shape=", dim(data_train$rep)))
    message("4. init train *.M, *.C, *.rep done!")

    # Update pseudo-label for test data
    data_test <- update_testset_R_C_M_K(args, model, data_test, dataloader_test, data_train)

    # Use the kmeans label
    test_final_embed <- data_test$rep
    test_final_embed <- as_array(test_final_embed)

    # Extract cluster centers from the trained kmeans model
    centers <- kmeans$centers
    # Calculate the nearest center for each row in test_final_embed
    # Define the function to assign each point to the nearest cluster
    assign_to_clusters <- function(new_data, centers) {
      apply(new_data, 1, function(row) {
        distances <- apply(centers, 1, function(center) sum((row - center)^2))  # Calculate squared Euclidean distances
        which.min(distances)  # Get the index of the closest cluster center (1 or 2)
      })
    }

    # Predict cluster membership for each point in test_final_embed_array
    test_cluster_old_labels <- assign_to_clusters(test_final_embed, centers)
    # Display the predicted cluster labels
    #print(test_cluster_old_labels)

    # R does not have a function to predict kmeans cluster membership. Python uses euclidean distance to make this prediction
    #test_cluster_old_labels <- predict(kmeans, newdata = test_final_embed)
    test_list_c <- test_cluster_old_labels - 1
    test_new_list_c <- sapply(test_list_c, function(x) order_c_map[as.character(x)])
    #data_test$C <- torch::torch_tensor(test_new_list_c)
    data_test$pred_C <- torch::torch_tensor(test_new_list_c)

    # Classification and regression
    list_train_AE_loss <- c()
    list_train_classifier_loss <- c()
    list_train_outcome_loss <- c()
    list_train_p_value_loss <- c()
    list_train_p_value_max <- c()
    list_train_p_value_min <- c()
    list_outcome_likelihood <- c()

    for (epoch in seq_len(args$epoch_in_iter)) {
      message(paste("epoch=", epoch))
      message("---------------------------------------------------------------------------------")
      message(paste("iter_i = ", iter_i, ", epoch=", epoch))

      total <- 0
      correct <- 0
      error_AE <- c()
      error_classifier <- c()
      error_outcome <- c()
      error_p_value <- c()
      error_outcome_likelihood <- c()
      outcome_true_y <- c()
      outcome_pred_prob <- c()
      message("-----------------")
      model$train()

      #dataloader_train$reset()
      current_index <- 1
      batch_iter <- dataloader_train$.iter()  # Create an iterator

      while (current_index <= dataloader_train$.length()){
        batch  <- batch_iter$.next()  # Get the first batch
        index <- batch[[1]]
        batch_xvy <- batch[[2]]
        batch_c <- batch[[3]]

        data_x <- batch[[1]]
        data_v <- batch[[2]]
        target <- batch[[3]]

        data_x <- torch::torch_tensor(data_x, requires_grad = FALSE)
        data_v <- torch::torch_tensor(data_v, requires_grad = FALSE)
        target <- torch::torch_tensor(target, requires_grad = FALSE)
        batch_c <- torch::torch_tensor(batch_c, requires_grad = FALSE)

        if (args$cuda) {
          data_x <- data_x$cuda()
          data_v <- data_v$cuda()
          target <- target$cuda()
          batch_c <- batch_c$cuda()
        }

        # Assuming batch size is 1 and sequence length is 1 for demonstration
        data_x <- data_x$unsqueeze(1)  # Add a dimension to make the shape [1, 1, 4]
        data_v <- data_v$unsqueeze(1)  # Add a dimension to make the shape [1, 1, 4]

        # x, function, demov, mask_BoolTensor
        output <- model$forward(x = data_x, function_name = "outcome_logistic_regression", demov = data_v)
        encoded_x <- output[[1]]
        decoded_x <- output[[2]]
        output_c_no_activate <- output[[3]]
        output_outcome <- output[[4]]

        #############################
        # For calculate p-value, mask k1, or, mask k1 and k2 together, to calculate G
        list_k_in_c <- seq_len(args$K_clusters)
        k1 <- sample(list_k_in_c, 1)
        k2 <- sample(list_k_in_c, 1)
        while (k2 == k1) {
          k2 <- sample(list_k_in_c, 1)
        }

        list_mask_k1 <- rep(0, args$K_clusters)
        list_mask_k1[k1] <- 1

        list_mask_k1k2 <- rep(0, args$K_clusters)
        list_mask_k1k2[c(k1, k2)] <- 1

        mask_k1_tensor <- torch::torch_tensor(list_mask_k1, dtype = torch_bool())
        mask_k1k2_tensor <- torch::torch_tensor(list_mask_k1k2, dtype = torch_bool())

        output_mask_k1 <- model$forward(x = data_x, function_name = "outcome_logistic_regression", demov = data_v, mask_BoolTensor = mask_k1_tensor)
        output_mask_k1k2 <- model$forward(x = data_x, function_name = "outcome_logistic_regression", demov = data_v, mask_BoolTensor = mask_k1k2_tensor)

        encoded_x_mask_k1 <- output_mask_k1[[1]]
        decoded_x_mask_k1 <- output_mask_k1[[2]]
        output_c_no_activate_mask_k1 <- output_mask_k1[[3]]
        output_outcome_mask_k1 <- output_mask_k1[[4]]

        encoded_x_mask_k1k2 <- output_mask_k1k2[[1]]
        decoded_x_mask_k1k2 <- output_mask_k1k2[[2]]
        output_c_no_activate_mask_k1k2 <- output_mask_k1k2[[3]]
        output_outcome_mask_k1k2 <- output_mask_k1k2[[4]]
        ###############################

        optimizer$zero_grad()

        # Increment labels to start from 1
        batch_c <- batch_c + 1
        # Ensure batch_c is in LongType for criterion_CrossEntropy
        batch_c <- batch_c$to(dtype = torch_long())
        # Squeeze the extra dimension to get the shape [1, 2]
        output_c_no_activate <- output_c_no_activate$squeeze(1)

        loss_classifier <- criterion_CrossEntropy(output_c_no_activate, batch_c)
        loss_AE <- criterion_MSE(data_x, decoded_x)
        loss_outcome <- criterion_BCE(output_outcome, target$float())
        loss_outcome_mask_k1 <- criterion_BCE(output_outcome_mask_k1, target$float())
        loss_outcome_mask_k1k2 <- criterion_BCE(output_outcome_mask_k1k2, target$float())
        loss_G <- 2 * (loss_outcome_mask_k1k2 - loss_outcome_mask_k1)
        loss_p_value <- 3.841 - loss_G

        loss <- args$lambda_AE * loss_AE +
          args$lambda_classifier * loss_classifier +
          args$lambda_outcome * loss_outcome +
          args$lambda_p_value * loss_p_value
        loss$backward()
        optimizer$step()

        error_AE <- c(error_AE, as.numeric(loss_AE$item()))
        error_classifier <- c(error_classifier, as.numeric(loss_classifier$item()))
        error_outcome <- c(error_outcome, as.numeric(loss_outcome$item()))
        error_p_value <- c(error_p_value, as.numeric(loss_p_value$item()))
        error_outcome_likelihood <- c(error_outcome_likelihood, as.numeric(loss_outcome$item()))

        # Get both values and indices by applying torch_max correctly
        max_values_indices <- torch_max(output_c_no_activate, dim = 2, keepdim = TRUE)
        # Extract the maximum values and the indices
        max_values <- max_values_indices[[1]]
        indices <- max_values_indices[[2]]

        predicted <- indices

        data_train$pred_C[current_index] <- as.numeric(as_array(torch::torch_tensor(predicted, dtype = torch_long())))
        total <- total + batch_c$size(1)
        correct <- correct + sum(as.numeric(predicted == batch_c))

        outcome_true_y <- c(outcome_true_y, as.numeric(target$data()))
        outcome_pred_prob <- c(outcome_pred_prob, as.numeric(output_outcome$data()))

        current_index <- current_index + 1

      }

      train_outcome_auc_score <- pROC::auc(outcome_true_y, outcome_pred_prob)
      message(paste("total=", total))
      classifier_c_accuracy <- correct / total

      train_AE_loss <- mean(error_AE)
      train_classifier_loss <- mean(error_classifier)
      train_outcome_loss <- mean(error_outcome)
      train_p_value_loss <- mean(error_p_value)
      train_outcome_likeilhood <- mean(error_outcome_likelihood)

      list_train_AE_loss <- c(list_train_AE_loss, train_AE_loss)
      list_train_classifier_loss <- c(list_train_classifier_loss, train_classifier_loss)
      list_train_outcome_loss <- c(list_train_outcome_loss, train_outcome_loss)

      list_train_p_value_max <- c(list_train_p_value_max, max(error_p_value))
      list_train_p_value_min <- c(list_train_p_value_min, min(error_p_value))
      list_train_p_value_loss <- c(list_train_p_value_loss, train_p_value_loss)

      list_outcome_likelihood <- c(list_outcome_likelihood, train_outcome_likeilhood)

      test_results <- suppressWarnings(func_analysis_test_error_D0406(args, model, data_test, dataloader_test))
      test_AE_loss <- test_results$test_AE_loss
      test_classifier_c_accuracy <- test_results$test_classifier_c_accuracy
      test_outcome_likelihood <- test_results$test_outcome_likelihood
      test_outcome_auc_score <- test_results$outcome_auc_score

      iter_train_negloglikeli_list <- c(iter_train_negloglikeli_list, train_outcome_likeilhood)
      iter_test_negloglikeli_list <- c(iter_test_negloglikeli_list, test_outcome_likelihood)
      iter_train_classifier_acc_list <- c(iter_train_classifier_acc_list, classifier_c_accuracy)
      iter_test_classifier_acc_list <- c(iter_test_classifier_acc_list, test_classifier_c_accuracy)
      iter_train_AE_list <- c(iter_train_AE_list, train_AE_loss)
      iter_test_AE_list <- c(iter_test_AE_list, test_AE_loss)
      iter_train_auc_list <- c(iter_train_auc_list, train_outcome_auc_score)
      iter_test_auc_list <- c(iter_test_auc_list, test_outcome_auc_score)

      message(sprintf("epoch %2d: train AE loss= %.4e, c acc= %.4e, outcome nll= %.4e, outcome_auc_score= %.4e, classifier loss= %.4e, outcome loss= %.4e, p_value loss= %.4e,",
                    epoch, train_AE_loss, classifier_c_accuracy, train_outcome_likeilhood, train_outcome_auc_score, train_classifier_loss, train_outcome_loss, train_p_value_loss))
      message(sprintf("        : test  AE loss= %.4e, c acc= %.4e, outcome nll= %.4e, outcome_auc_score= %.4e",
                    test_AE_loss, test_classifier_c_accuracy, test_outcome_likelihood, test_outcome_auc_score))

      analysis_out  <- analysis_p_value_related(data_train, args$K_clusters, 1)
      dict_p_value  <- analysis_out$dict_p_value
      dict_p_value_list <- unlist(dict_p_value)
      flag_morethan_0p05 <- any(dict_p_value_list > 0.05)

      last_p <- as.numeric(dict_p_value[[length(dict_p_value)]])
      message(paste("test_outcome_likelihood=", test_outcome_likelihood))
      message(paste("min_test_negloglikeli_record=", min_test_negloglikeli_record))
      message(test_outcome_likelihood <= min_test_negloglikeli_record)
      message(length(dict_p_value))
      message(paste(names(dict_p_value), unlist(dict_p_value), sep = "=", collapse = ", "))
      message(paste("last_p_value=", last_p))

      if ((test_outcome_likelihood <= min_test_negloglikeli_record) &
          !is.na(last_p) & last_p <= 0.05) {
        message(paste("save model here! iter_i=", iter_i, ", epoch=", epoch))
        min_test_negloglikeli_record <- test_outcome_likelihood
        torch_save(model$state_dict(), file.path(part2_foldername, 'model_iter.pt'))
        message("    Saving model")

        data_train$M <- as.array(data_train$M)
        data_train$C <- as.numeric(as.array(data_train$C))
        data_train$pred_C <- as.numeric(as.array(data_train$pred_C))
        data_train$pred_C <- data_train$pred_C - 1
        saveRDS(data_train, file = file.path(part2_foldername, 'data_train_iter.rds'))
        message("    save data_train")
        data_test$M <- as.array(data_test$M)
        #data_test$C <- as.array(data_test$C)
        #data_test$pred_C <- as.array(data_test$pred_C)
        data_test$C <- as.numeric(as.array(data_test$C))
        data_test$pred_C <- as.numeric(as.array(data_test$pred_C))
        #data_test$pred_C <- data_test$pred_C - 3
        saveRDS(data_test, file = file.path(part2_foldername, 'data_test_iter.rds'))
        message("    save data_test")

        saved_iter_list <- c(saved_iter_list, iter_i)
        saved_iter <- iter_i
      }
    }


  }

  message(paste("number_reassign_list=", number_reassign_list))

  min_iter_test_negloglikel <- min(iter_test_negloglikeli_list)
  index_min_iter_test_negloglikeli <- which.min(iter_test_negloglikeli_list)
  number_reassign_list_tosee <- number_reassign_list[seq_len(index_min_iter_test_negloglikeli)]
  train_negloglikeli_tosee <- iter_train_negloglikeli_list[index_min_iter_test_negloglikeli]
  classifier_acc_tosee <- iter_train_classifier_acc_list[index_min_iter_test_negloglikeli]

  message(paste("random_state_list=", random_state_list))
  message(paste("saved_iter_list=", saved_iter_list))
  message(paste("saved_iter = ", saved_iter))
}

# Placeholder for the yf_dataset_withdemo class
yf_dataset_withdemo <- function(path, file_name, n_z) {
  data <- readRDS(paste0(path, file_name))
  n_samp <- nrow(data[[1]])
  dataset <- list(
    data_x = data[[1]],
    data_v = data[[2]],
    data_y = data[[3]],
    n_samples = n_samp,
    n_cat = NULL,
    M = NULL,
    C = torch::torch_tensor(rep(0L, n_samp), dtype = torch_long()),
    pred_C = torch::torch_tensor(rep(0L, n_samp), dtype = torch_long()),
    rep = NULL
  )
  return(dataset)
}

#' Parse Command-Line Arguments for PPD-Aware Clustering
#'
#' This function parses command-line arguments for a script that performs PPD-aware clustering, setting up various parameters necessary for training and evaluating a model.
#'
#' @return A list of parsed command-line arguments, each corresponding to a specific parameter required by the clustering model. This list is typically used to configure the model training and evaluation process.
#'
#' @details
#' The \code{parse_args} function uses the \code{argparser} package to define and parse a series of command-line arguments. These arguments are essential for configuring various aspects of the PPD-aware clustering process, including model architecture, training settings, and input/output paths.
#'
#' The following command-line arguments are supported:
#' \itemize{
#'   \item \code{--init_AE_epoch}: Integer. Number of epochs for representation initialization.
#'   \item \code{--n_hidden_fea}: Integer. Number of hidden units in the LSTM layers.
#'   \item \code{--output_path}: Character. Location where output results will be saved.
#'   \item \code{--input_path}: Character. Location of the input dataset.
#'   \item \code{--filename_train}: Character. Path to the training data file.
#'   \item \code{--filename_test}: Character. Path to the test data file.
#'   \item \code{--n_input_fea}: Integer. Number of original input features.
#'   \item \code{--n_dummy_demov_fea}: Integer. Number of dummy demographic features.
#'   \item \code{--lstm_layer}: Integer. Number of layers in the LSTM. Default is 1.
#'   \item \code{--lr}: Numeric. Learning rate for training. Default is 1e-4.
#'   \item \code{--lstm_dropout}: Numeric. Dropout rate for the LSTM layers. Default is 0.0.
#'   \item \code{--K_clusters}: Integer. Number of initial clusters.
#'   \item \code{--iter}: Integer. Maximum number of iterations for merging clusters. Default is 20.
#'   \item \code{--epoch_in_iter}: Integer. Number of epochs per iteration when merging clusters. Default is 1.
#'   \item \code{--seed}: Integer. Random seed for reproducibility. Default is 1111.
#'   \item \code{--cuda}: Integer. Flag indicating whether to use CUDA (GPU acceleration). Default is 1.
#'   \item \code{--lambda_AE}: Numeric. Regularization weight for the autoencoder loss in each iteration. Default is 1.0.
#'   \item \code{--lambda_classifier}: Numeric. Regularization weight for the classifier loss in each iteration. Default is 1.0.
#'   \item \code{--lambda_outcome}: Numeric. Regularization weight for the outcome loss in each iteration. Default is 1.0.
#'   \item \code{--lambda_p_value}: Numeric. Regularization weight for the p-value adjustment in each iteration. Default is 1.0.
#' }
#' @examples
#' \donttest{
#' ## parse_args() reads from the command line; use the args list interface
#' ## directly when calling DICEr() from within R (see ?DICEr).
#' args <- parse_args()
#' }
#' @import argparser
#' @export


parse_args <- function() {
  # Defines the function parse_args() which doesn't accept any parameters and is intended to be called directly from the command line or within a script to parse arguments.
  parser <- arg_parser('ppd-aware clustering')

  # Adding Arguments
  # Each add_argument() call adds a specific command-line option to the parser. Here is what each one does:
  parser <- add_argument(parser, '--init_AE_epoch', type = "integer", help = 'number of epoch for representation initialization'#, required = TRUE
  )
  parser <- add_argument(parser, '--n_hidden_fea', type = "integer", help = 'number of hidden size in LSTM'#, required = TRUE
  )
  parser <- add_argument(parser, '--output_path', type = "character", help = 'location of output path')
  parser <- add_argument(parser, '--input_path', type = "character", help = 'location of input dataset'#, required = TRUE
  )
  parser <- add_argument(parser, '--filename_train', type = "character", help = 'location of the data corpus'#, required = TRUE
  )
  parser <- add_argument(parser, '--filename_test', type = "character", help = 'location of the data corpus'#, required = TRUE
  )
  parser <- add_argument(parser, '--n_input_fea', type = "integer", help = 'number of original input feature size'#, required = TRUE
  )
  parser <- add_argument(parser, '--n_dummy_demov_fea', type = "integer", help = 'number of dummy demo feature size'#, required = TRUE
  )
  parser <- add_argument(parser, '--lstm_layer', type = "integer", help = 'number of hidden size in LSTM', default = 1)
  parser <- add_argument(parser, '--lr', type = "numeric", help = 'learning rate', default = 1e-4)
  parser <- add_argument(parser, '--lstm_dropout', type = "numeric", help = 'dropout in LSTM', default = 0.0)
  parser <- add_argument(parser, '--K_clusters', type = "integer", help = 'number of initial clusters'#, required = TRUE
  )
  parser <- add_argument(parser, '--iter', type = "integer", help = 'maximum of iterations in iteration merge clusters', default = 20)
  parser <- add_argument(parser, '--epoch_in_iter', type = "integer", help = 'maximum of iterations in iteration merge clusters', default = 1)
  parser <- add_argument(parser, '--seed', type = "integer", help = 'random seed', default = 1111)
  parser <- add_argument(parser, '--cuda', type = "integer", help = 'If use cuda', default = 1)
  parser <- add_argument(parser, '--lambda_AE', type = "numeric", help = 'lambda of AE in iteration', default = 1.0)
  parser <- add_argument(parser, '--lambda_classifier', type = "numeric", help = 'lambda_classifier of classifier in iteration', default = 1.0)
  parser <- add_argument(parser, '--lambda_outcome', type = "numeric", help = 'lambda of outcome in iteration', default = 1.0)
  parser <- add_argument(parser, '--lambda_p_value', type = "numeric", help = 'lambda of p value in iteration', default = 1.0)

  # Parse & display arguments
  # - parser.parse_args(): Parses the command-line arguments provided to the script. If any required arguments are missing or invalid, the script will automatically display an error and help information.
  # - print(vars(args)): Converts the args Namespace object to a dictionary and prints it. This shows all the arguments that have been set, either by defaults or by user input.
  # - return args: Returns the populated args Namespace. This object contains all the command-line arguments that were parsed. This is useful for the rest of the program to access and use these settings.
  args <- argparser::parse_args(parser)
  message("parameters:")
  message(paste(utils::capture.output(args), collapse = "\n"))
  return(args)
}

#' Calculate P-Value for Logistic Regression Model
#'
#' This function calculates the p-value for a logistic regression model by comparing the likelihoods of a full model and a null model.
#'
#' @param X A matrix of predictor variables for the full model.
#' @param y A binary response vector, with values of 0 or 1.
#' @param is_intercept A logical value indicating whether the null model should include only an intercept (i.e., no predictors). If \code{TRUE}, the null model includes only the intercept; if \code{FALSE}, the null model is specified by \code{X_null}.
#' @param X_null (Optional) A matrix of predictor variables for the null model. This argument is required if \code{is_intercept} is \code{FALSE}.
#'
#' @return A numeric value representing the p-value from the likelihood ratio test between the full and null models.
#'
#' @details
#' The \code{p_value_calculate} function performs the following steps:
#' \enumerate{
#'   \item Fits a full logistic regression model using \code{X} as the predictor matrix.
#'   \item Depending on the value of \code{is_intercept}, it either:
#'   \itemize{
#'     \item Calculates the null model's log-likelihood using only an intercept (if \code{is_intercept = TRUE}).
#'     \item Fits a null logistic regression model using \code{X_null} as the predictor matrix (if \code{is_intercept = FALSE}).
#'   }
#'   \item Computes the likelihood ratio statistic \code{G}, which is twice the difference in log-likelihoods between the full and null models.
#'   \item Calculates the p-value using the chi-squared distribution with degrees of freedom equal to the difference in the number of parameters between the full and null models.
#' }
#'
#' @examples
#' # Example with only an intercept in the null model
#' X <- matrix(rnorm(100 * 5), ncol = 5)
#' y <- rbinom(100, 1, 0.5)
#' p_value <- p_value_calculate(X, y, is_intercept = TRUE)
#' print(p_value)
#'
#' # Example with a specified null model
#' X_null <- X[, 1:2]  # Using only the first two predictors in the null model
#' p_value <- p_value_calculate(X, y, is_intercept = FALSE, X_null = X_null)
#' print(p_value)
#'
#' @import stats
#' @export


p_value_calculate <- function(X, y, is_intercept, X_null = NULL) {
  message("X.shape=", dim(X), ", y.shape=", length(y), "")

  # Fit the full model
  full_model <- stats::glm(y ~ X, family = binomial())
  alt_log_likelihood <- stats::logLik(full_model)

  if (isTRUE(as.logical(is_intercept))) {
    # Null model = intercept only
    null_prob <- mean(y)
    null_log_likelihood <- sum(stats::dbinom(y, size = 1, prob = null_prob, log = TRUE))
    df <- 1
    G <- 2 * (alt_log_likelihood - null_log_likelihood)
    p_value <- stats::pchisq(G, df, lower.tail = FALSE)
  } else {
    # Null model specified by X_null
    null_model <- stats::glm(y ~ X_null, family = binomial())
    null_log_likelihood <- stats::logLik(null_model)
    df <- max(1L, ncol(X) - ncol(X_null))
    G <- 2 * (alt_log_likelihood - null_log_likelihood)
    p_value <- stats::pchisq(G, df, lower.tail = FALSE)
  }
  return(p_value)
}

#' Update Cluster Embeddings in Matrix M
#'
#' This function updates the matrix \code{M}, which contains cluster embeddings, based on the current cluster assignments and representations in the training data.
#'
#' @param data_train A list containing the training data. The list must include the following components:
#' \itemize{
#'   \item \code{M}: A matrix where each column corresponds to a cluster's embedding vector.
#'   \item \code{n_cat}: An integer representing the number of clusters (or categories).
#'   \item \code{rep}: A matrix of representations (embeddings) for the training data.
#'   \item \code{C}: An integer vector indicating the cluster assignment for each training sample.
#' }
#'
#' @return The function updates the \code{M} matrix in-place within the \code{data_train} list, reflecting the new average embeddings for each cluster.
#'
#' @details
#' The \code{update_M} function works as follows:
#' \enumerate{
#'   \item It first checks that the number of columns in the \code{M} matrix matches the number of clusters (\code{n_cat}). If not, an error is thrown.
#'   \item It then initializes a list of lists, \code{dict_c_embedding}, where each sublist will hold the embeddings for samples belonging to a specific cluster.
#'   \item The function iterates over all samples, adding their representations to the corresponding cluster sublist in \code{dict_c_embedding}.
#'   \item For each cluster, the function computes the mean of the embeddings across all samples assigned to that cluster.
#'   \item The computed mean embedding is then used to update the corresponding column in the \code{M} matrix.
#' }
#' @examples
#' \donttest{
#' ## Called internally by DICEr() to refresh cluster centroids each iteration.
#' ## Requires a data_train list with rep, C, n_cat, and M fields; see ?DICEr.
#' }
#' @import torch
#' @export

update_M <- function(data_train) {
  # Check if the number of columns in M matches n_cat
  if (ncol(data_train$M) != data_train$n_cat) {
    stop("Invalid M!")
  }

  dict_c_embedding <- vector("list", data_train$n_cat)
  representations <- data_train$rep
  list_C <- as.integer(data_train$C)

  for (i in seq_len(data_train$n_cat)) {
    dict_c_embedding[[i]] <- list()
  }

  for (i in seq_along(list_C)) {
    cp <- list_C[i]
    dict_c_embedding[[cp + 1]] <- c(dict_c_embedding[[cp + 1]], list(representations[i, , drop = FALSE]))
  }

  for (i in seq_len(data_train$n_cat)) {
    c_key <- i
    if (length(dict_c_embedding[[c_key]]) == 0) {
      next
    }
    embed_list <- torch_stack(dict_c_embedding[[c_key]], dim = 1)
    embed_mean_dim0 <- torch_mean(embed_list, dim = 2)

    # Calculate the mean along dimension 1 (rows), keeping only 4 values (one for each column)
    mean_values <- torch_mean(embed_mean_dim0, dim = 1)

    data_train$M[, c_key] <- mean_values
  }
  message("    update M!")
}

#' Analyze P-Value and Outcome Ratios for Clusters
#'
#' This function performs an analysis of p-values and outcome ratios related to clusters in a dataset. It calculates the ratio of positive outcomes within each cluster and computes p-values for the association between clusters and the outcome variable.
#'
#' @param data_train A list representing the training data. This list must include the following components:
#' \itemize{
#'   \item \code{C}: An integer vector of cluster assignments for each sample.
#'   \item \code{data_v}: A matrix or data frame of additional covariates.
#'   \item \code{data_y}: A binary response vector with values 0 or 1, representing the outcome variable.
#' }
#' @param num_clusters An integer specifying the number of clusters in the data.
#' @param if_check A logical value indicating whether to print intermediate steps for debugging purposes. Default is \code{FALSE}.
#'
#' @return A list containing the following components:
#' \itemize{
#'   \item \code{dict_outcome_ratio}: A numeric vector representing the ratio of positive outcomes (where \code{data_y} equals 1) within each cluster.
#'   \item \code{dict_p_value}: A list of p-values, with each element corresponding to a pair of clusters. The p-values are calculated using a likelihood ratio test for the association between the clusters and the outcome variable.
#' }
#'
#' @details
#' The \code{analysis_p_value_related} function performs the following steps:
#' \enumerate{
#'   \item Converts the cluster assignments into a one-hot encoded matrix and counts the occurrences of each cluster.
#'   \item Calculates the ratio of positive outcomes within each cluster.
#'   \item Combines the cluster one-hot encodings with additional covariates to create a matrix of predictor variables.
#'   \item Iteratively removes one or two clusters from the predictor matrix and calculates the p-value for the remaining clusters using the \code{\link{p_value_calculate}} function.
#' }
#' @examples
#' set.seed(1)
#' n <- 80L
#' data_train <- list(
#'   C      = sample(0:1, n, replace = TRUE),
#'   data_v = matrix(rbinom(n * 3L, 1, 0.5), n, 3L),
#'   data_y = rbinom(n, 1, 0.3)
#' )
#' result <- analysis_p_value_related(data_train, num_clusters = 2L)
#' result$dict_outcome_ratio   # outcome rate per cluster
#' result$dict_p_value          # LRT p-value for the cluster pair
#' @export


analysis_p_value_related <- function(data_train, num_clusters, if_check = FALSE) {
  data_C <- data_train$C
  data_v <- data_train$data_v
  data_y <- data_train$data_y

  list_c <- as.integer(data_C)
  list_onehot <- matrix(0, nrow = length(list_c), ncol = num_clusters)
  dict_c_count <- integer(num_clusters)
  dict_outcome_in_c_count <- integer(num_clusters)

  for (i in seq_along(list_c)) {
    list_onehot[i, list_c[i] + 1] <- 1
    dict_c_count[list_c[i] + 1] <- dict_c_count[list_c[i] + 1] + 1
    if (data_y[i] == 1) {
      dict_outcome_in_c_count[list_c[i] + 1] <- dict_outcome_in_c_count[list_c[i] + 1] + 1
    }
  }

  if (if_check) {
    message("--------")
    message("num_clusters=", num_clusters, "")
    message("")
    message("list_c[0]=", list_c[1], "")
    message("list_onehot[0,]=", list_onehot[1,], "")
    message("")
    message("list_c[1]=", list_c[2], "")
    message("list_onehot[1,]=", list_onehot[2,], "")
    message("--------")
  }
  message("dict_c_count=", dict_c_count, "")
  message("dict_outcome_in_c_count=", dict_outcome_in_c_count, "")

  dict_outcome_ratio <- dict_outcome_in_c_count / dict_c_count
  message("dict_outcome_ratio=", dict_outcome_ratio, "")

  var_c <- list_onehot
  var_v <- data_v
  depend_y <- data_y

  var_cpv <- cbind(var_c, var_v)
  if (if_check) {
    message("var_c.shape=", dim(var_c), ", var_v.shape=", dim(var_v), ", depend_y.shape=", length(depend_y), "")
    message("var_cpv.shape=", dim(var_cpv), "")
  }

  message("analysis done!")
  dict_p_value <- list()
  for (k1 in seq_len(num_clusters) - 1) {
    X_remove_k1 <- var_cpv
    slices_k1 <- setdiff(seq_len(ncol(var_cpv)), k1 + 1)
    X_remove_k1 <- X_remove_k1[, slices_k1, drop = FALSE]

    for (k2 in seq_len(max(0L, num_clusters - k1 - 1L)) + k1 + 1L) {
      X_remove_k1k2 <- var_cpv
      slices_k1k2 <- setdiff(seq_len(ncol(var_cpv)), c(k1 + 1, k2))
      X_remove_k1k2 <- X_remove_k1k2[, slices_k1k2, drop = FALSE]

      message("---------")
      message("k1=", k1, ", k2=", k2 - 1, "")
      message("slices_k1=", slices_k1, "")
      message("slices_k1k2=", slices_k1k2, "")
      p_value_k1k2 <- p_value_calculate(X_remove_k1, depend_y, 0, X_remove_k1k2)
      dict_p_value[[paste(k1, k2 - 1, sep = ",")]] <- p_value_k1k2
    }
  }
  message("dict_p_value=", unlist(dict_p_value), "")
  return(list(dict_outcome_ratio = dict_outcome_ratio, dict_p_value = dict_p_value))
}

#' EncoderRNN Neural Network Module
#'
#' An RNN-based encoder module using Long Short-Term Memory (LSTM) architecture, implemented with the \code{torch} package.
#' This module is designed to process sequential data and is commonly used in tasks like sequence-to-sequence modeling, language modeling, and time series forecasting.
#'
#' @section Usage:
#' \preformatted{
#' EncoderRNN <- nn_module(
#'   "EncoderRNN",
#'
#'   initialize = function(input_size, nhidden, nlayers, dropout) {
#'     ...
#'   },
#'
#'   forward = function(x) {
#'     ...
#'   }
#' )
#' }
#'
#' @section Methods:
#' \describe{
#'   \item{\code{initialize(input_size, nhidden, nlayers, dropout)}}{
#'     Initializes the EncoderRNN module.
#'
#'     \itemize{
#'       \item{\code{input_size}: A numeric value representing the number of features in the input data.}
#'       \item{\code{nhidden}: A numeric value representing the number of hidden units in each LSTM layer.}
#'       \item{\code{nlayers}: A numeric value indicating the number of LSTM layers.}
#'       \item{\code{dropout}: A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.}
#'     }
#'   }
#'
#'   \item{\code{forward(x)}}{
#'     Executes the forward pass of the EncoderRNN module.
#'
#'     \itemize{
#'       \item{\code{x}: A tensor of shape (batch_size, sequence_length, input_size) representing the input data.}
#'     }
#'
#'     \value{
#'     A list containing:
#'       \itemize{
#'         \item{\code{output}: The output tensor from the LSTM, processed and flipped.}
#'         \item{\code{list(hn, cn)}: A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM.}
#'         \item{\code{newinput}: A tensor representing the input data with an additional zero tensor at the beginning.}
#'       }
#'     }
#'   }
#' }
#'
#' @param input_size A numeric value representing the number of features in the input data.
#' @param nhidden A numeric value representing the number of hidden units in each LSTM layer.
#' @param nlayers A numeric value indicating the number of LSTM layers.
#' @param dropout A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.
# @param x A tensor of shape \code{(batch_size, sequence_length, input_size)} representing the input data.
# @param h A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM, typically passed from the encoder or previous decoder step.

#' @return An \code{nn_module} object (torch LSTM-based encoder). When called
#'   with a tensor of shape \code{(batch_size, seq_len, input_size)}, returns a
#'   list of: (1) processed output tensor, (2) list of hidden and cell states
#'   \code{(hn, cn)}, and (3) the input with a leading zero tensor prepended.
#' @examples
#' \donttest{
#' enc <- EncoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
#' x   <- torch_randn(2L, 1L, 4L)   # batch=2, seq_len=1, features=4
#' out <- enc(x)
#' dim(out[[1]])  # encoded output
#' }
#' @import torch
#' @export




EncoderRNN <- torch::nn_module(
  "EncoderRNN",

  initialize = function(input_size, nhidden, nlayers, dropout) {
    self$nhidden <- nhidden
    self$feasize <- input_size
    self$nlayers <- nlayers
    self$dropout <- dropout

    # Define the LSTM layer
    self$lstm <- nn_lstm(
      input_size = input_size, #self$feasize,
      hidden_size = nhidden, #self$nhidden,
      num_layers = nlayers, #self$nlayers,
      dropout = dropout, #self$dropout,
      batch_first = TRUE
    )

    # Initialize weights directly
    #params <- self$lstm$parameters()
    #for (p in params) {
    #  p$data$uniform_(-0.1, 0.1)
    #}
  },

  forward = function(x) {
    batch_size <- x$size(1)
    # Forward pass through LSTM
    out_state <- self$lstm(x)
    output <- out_state[[1]]
    state <- out_state[[2]]
    hn <- state[[1]]
    cn <- state[[2]]

    # Process the output
    output <- torch_flip(output, dims = c(1))
    newinput <- torch_flip(x, dims = c(1))

    # Add a zero tensor
    zeros <- torch_zeros(c(batch_size, 1, x$size(3)))
    newinput <- torch_cat(list(zeros, newinput), 2)
    #newinput <- newinput$narrow(2, 0, newinput$size(2) - 1)
    #newinput <- torch::torch_reshape(newinput, c(batch_size, sequence_length, input_size))
    #newinput <- torch::torch_reshape(newinput, c(batch_size, sequence_length, input_size))

    # Return the processed output and new input
    list(output, list(hn, cn), newinput)
  }

)

#' Evaluate Autoencoder Model on Test Data
#'
#' This function evaluates the performance of an autoencoder model on a test dataset, calculating the mean squared error (MSE) between the original inputs and the reconstructed outputs.
#'
#' @param args A list of arguments, typically containing configurations such as whether to use CUDA for computation.
#' @param model A trained autoencoder model that will be evaluated. The model is expected to have a `forward` method that takes input data and a mode (e.g., "autoencoder").
#' @param dataloader_test A list representing the test data loader. Each element in the list should be a batch containing test data that the model will evaluate.
#'
#' @return A numeric value representing the mean test error (MSE) across all batches in the test dataset.
#'
#' @details
#' The \code{test_AE} function performs the following steps:
#' \enumerate{
#'   \item Initializes the mean squared error (MSE) loss criterion.
#'   \item Sets the model to evaluation mode to ensure layers like dropout are appropriately handled.
#'   \item Iterates over the test data batches provided by \code{dataloader_test}.
#'   \item For each batch, the function:
#'   \itemize{
#'     \item Extracts the data, ensuring it is in numeric format and reshapes it as necessary.
#'     \item Converts the data into torch tensors, moving them to CUDA if specified.
#'     \item Performs a forward pass through the model in autoencoder mode, obtaining both the encoded representation and the reconstructed output.
#'     \item Computes the MSE loss between the input and the reconstructed output.
#'     \item Accumulates the loss for each batch.
#'   }
#'   \item Computes the mean of the accumulated losses to return the overall test error.
#' }
#' @examples
#' \donttest{
#' ## Called internally by DICEr() after each autoencoder warm-up epoch.
#' ## Requires a trained EncoderRNN/DecoderRNN model; see ?DICEr.
#' }
#' @import torch
#' @export


test_AE <- function(args, model, dataloader_test) {
  # Define the MSE loss criterion
  criterion_MSE <- nn_mse_loss()

  test_error <- c()
  message("-----------------")

  # Set the model to evaluation mode
  model$eval()

  suppressWarnings( {
    for (batch in 1:length(dataloader_test)) {
      # To get the batch data, use an iterator
      batch_iter <- dataloader_test$.iter()  # Create an iterator
      batch_idx  <- batch_iter$.next()  # Get the first batch
      data_x <- batch_idx[[1]]
      data_x <- torch_tensor(data_x, requires_grad = FALSE)
      data_x <- data_x$unsqueeze(1)

      # Perform a forward pass through the model in autoencoder mode
      output <- model$forward(data_x, "autoencoder")
      enc <- output[[1]]
      pred <- output[[2]]

      # Compute the loss
      loss <- criterion_MSE(data_x, pred)

      # Append the loss to the test_error list
      test_error <- c(test_error, as.numeric(loss$item()))
    }
  } )

  # Compute the mean test error
  test_AE_error <- mean(test_error)
  return(test_AE_error)
}

---
title: "DICErClust Illustrated Example: Heart Failure Risk Stratification"
author: "Sarah Ayton and Yiye Zhang"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{DICErClust Illustrated Example: Heart Failure Risk Stratification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
## All computational chunks are skipped on CRAN: this vignette downloads
## external data and runs DICEr() (~15-20 min on CPU), both of which are
## incompatible with CRAN's check environment.
## devtools::check() sets NOT_CRAN=true automatically for local builds.
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  fig.width  = 5,
  fig.height = 4,
  eval       = identical(Sys.getenv("NOT_CRAN"), "true")
)
```

## Overview

This vignette walks through a complete DICErClust workflow on the UCI Heart
Failure Clinical Records dataset (Chicco & Jurman 2020).  By the end you will
have:

- split 299 patients into training and test sets,
- trained DICEr to find two risk-stratified clusters,
- evaluated the clusters with an AUC and chi-squared test, and
- produced two publication-quality figures.

**Approximate run time** on a single CPU core: 15–20 minutes
(the training step is the bottleneck; all other steps are fast).

---

## 1. Load DICErClust

DICErClust is distributed as a source tarball.  Install it once with
`install.packages()`, then load it like any other package.

```{r load-pkg, eval = FALSE}
## Install from local tarball (run once):
# install.packages(
#   "/path/to/DICErClust_0.1.1.tar.gz",
#   repos = NULL, type = "source"
# )
library(DICErClust)
library(ggplot2)
library(pROC)
```

```{r load-pkg-real, include = FALSE}
## When building the vignette from within the package source tree we use
## devtools::load_all() so edits to the source are reflected immediately.
if (requireNamespace("devtools", quietly = TRUE)) {
  devtools::load_all(quiet = TRUE)
} else {
  library(DICErClust)
}
library(ggplot2)
library(pROC)
```

---

## 2. Download the UCI Heart Failure dataset

Chicco & Jurman (2020) collected 299 heart-failure patients from the
Faisalabad Institute of Cardiology in Pakistan.  The dataset contains
12 clinical features and a binary outcome `DEATH_EVENT` (1 = died during
follow-up; 0 = survived).  It is freely available from the UCI Machine
Learning Repository (dataset #519).

```{r download-data}
hf_url  <- paste0(
  "https://archive.ics.uci.edu/ml/",
  "machine-learning-databases/00519/",
  "heart_failure_clinical_records_dataset.csv"
)
hf_dest <- tempfile(fileext = ".csv")
download.file(hf_url, hf_dest, quiet = TRUE)
hf <- read.csv(hf_dest)

cat(sprintf("Rows: %d   Columns: %d\n", nrow(hf), ncol(hf)))
print(table(DEATH_EVENT = hf$DEATH_EVENT))
```

The dataset is small but class-imbalanced: roughly 68% of patients survived
(DEATH_EVENT = 0) and 32% died (DEATH_EVENT = 1).

---

## 3. Feature engineering

DICEr expects the input to be split into **two** feature matrices:

| Matrix | Contents | Role |
|--------|----------|------|
| `data_x` | Continuous laboratory / physiological measurements | Encoder input: the LSTM compresses these into a low-dimensional latent representation used for clustering |
| `data_v` | Binary demographic / comorbidity indicators | Auxiliary outcome-head input: the likelihood-ratio test assesses whether the cluster explains outcome **above and beyond** these covariates |

This design mirrors real-world EHR practice: `data_x` captures time-varying
lab values while `data_v` captures static patient characteristics.

```{r features}
## Continuous lab features → LSTM encoder (data_x)
x_cols <- c("age", "creatinine_phosphokinase", "ejection_fraction",
            "platelets", "serum_creatinine", "serum_sodium", "time")

## Binary demographic indicators → outcome head (data_v)
v_cols <- c("anaemia", "diabetes", "high_blood_pressure", "sex", "smoking")

## Min-max scale continuous features to [0, 1].
## Scaling prevents any single lab value from dominating the MSE
## reconstruction loss relative to others.
scale_01 <- function(x) {
  r <- range(x, na.rm = TRUE)
  if (diff(r) == 0) return(x * 0)
  (x - r[1]) / diff(r)
}

X_x <- apply(as.matrix(hf[, x_cols]), 2, scale_01)  # 299 × 7, numeric
X_v <- apply(as.matrix(hf[, v_cols]), 2, as.numeric) # 299 × 5, binary as float

## Note: data_v *must* be stored as numeric (float), not integer.
## torch_tensor() infers dtype from R storage mode; integer columns produce
## int64 tensors that are incompatible with the float32 model weights.

cat(sprintf("data_x: %d × %d\ndata_v: %d × %d\n",
            nrow(X_x), ncol(X_x), nrow(X_v), ncol(X_v)))

n_x <- ncol(X_x)  # 7  continuous predictors
n_v <- ncol(X_v)  # 5  binary demographics
outcome <- hf$DEATH_EVENT
```

---

## 4. Stratified train / test split

A 70 / 30 stratified split preserves the ~32% event rate in both partitions,
which is important given the small sample size.

```{r split}
set.seed(1111)
idx_death <- which(outcome == 1)
idx_alive <- which(outcome == 0)

train_idx <- sort(c(
  sample(idx_death, floor(0.70 * length(idx_death))),
  sample(idx_alive, floor(0.70 * length(idx_alive)))
))
test_idx <- setdiff(seq_len(nrow(hf)), train_idx)

cat(sprintf("Train: %d patients  (deaths: %d, %.0f%%)\n",
            length(train_idx), sum(outcome[train_idx]),
            100 * mean(outcome[train_idx])))
cat(sprintf("Test : %d patients  (deaths: %d, %.0f%%)\n",
            length(test_idx),  sum(outcome[test_idx]),
            100 * mean(outcome[test_idx])))
```

---

## 5. Serialise data in DICErClust format

`DICEr()` reads training and test sets from RDS files.  Each file must
contain a length-3 list:

1. `data_x` — numeric matrix, shape *n* × *p*
2. `data_v` — numeric matrix, shape *n* × *q*
3. `data_y` — integer vector of 0/1 outcomes, length *n*

```{r save-rds}
data_dir <- file.path(tempdir(), "dice_hf")
dir.create(data_dir, showWarnings = FALSE)

saveRDS(
  list(X_x[train_idx, ], X_v[train_idx, ], as.integer(outcome[train_idx])),
  file.path(data_dir, "hf_train.rds")
)
saveRDS(
  list(X_x[test_idx, ], X_v[test_idx, ], as.integer(outcome[test_idx])),
  file.path(data_dir, "hf_test.rds")
)
```

---

## 6. Configure DICEr

The argument list controls both the architecture and the training schedule.

```{r configure}
args <- list(
  seed              = 1111,          # reproducibility seed
  input_path        = data_dir,      # directory containing RDS files
  filename_train    = "hf_train.rds",
  filename_test     = "hf_test.rds",

  ## ── Architecture ──────────────────────────────────────────
  n_input_fea       = n_x,          # 7 continuous LSTM input features
  n_hidden_fea      = 4,            # LSTM latent dimension (7 → 4)
  lstm_layer        = 1,            # single LSTM layer
  lstm_dropout      = 0.0,          # no dropout (small dataset)
  K_clusters        = 2,            # binary risk partition: high vs. low

  ## ── Auxiliary features ────────────────────────────────────
  n_dummy_demov_fea = n_v,          # 5 binary demographic covariates

  ## ── Hardware ──────────────────────────────────────────────
  cuda              = FALSE,        # set TRUE for GPU acceleration

  ## ── Optimiser ─────────────────────────────────────────────
  lr                = 1e-4,         # Adam learning rate

  ## ── Training schedule ─────────────────────────────────────
  init_AE_epoch     = 5,            # Stage 1: autoencoder warm-up epochs
  iter              = 30,           # Stage 2: number of clustering iterations
  epoch_in_iter     = 2,            # gradient-update epochs per iteration

  ## ── Loss weights ──────────────────────────────────────────
  ## Combined loss: L = λ_AE·L_AE + λ_clf·L_classifier
  ##                  + λ_out·L_outcome + λ_p·L_p_value
  ## L_p_value = 3.841 − G penalises non-significant cluster configurations
  ## (G is the LRT statistic; 3.841 is the χ²(1) critical value at α = 0.05)
  lambda_AE         = 1.0,
  lambda_classifier = 1.0,
  lambda_outcome    = 1.0,
  lambda_p_value    = 1.0
)
```

### Training stages explained

DICEr runs three sequential stages:

1. **LSTM autoencoder warm-up** (`init_AE_epoch` epochs): trains the encoder
   and decoder to reconstruct `data_x`, establishing a compact latent
   representation before any clustering begins.

2. **Joint optimisation** (`iter` iterations × `epoch_in_iter` epochs each):
   alternates between
   (a) k-means clustering in the LSTM latent space, and
   (b) gradient updates minimising the combined four-component loss.
   After each k-means step the cluster with the highest `data_y = 1` rate is
   relabelled cluster 0 (the high-risk cluster).

3. **Model selection**: saves the checkpoint with the lowest test negative
   log-likelihood that *also* satisfies the likelihood-ratio test
   p < 0.05 between at least one cluster pair.

---

## 7. Train the model

```{r train, eval = FALSE}
## DICEr writes output files relative to the working directory.
## We temporarily switch to tempdir() to keep them self-contained.
old_wd <- setwd(tempdir())
suppressWarnings(DICEr(args))
setwd(old_wd)
```

Output is written to `hn_4_K_2/part2_AE_nhidden_4/` relative to the working
directory used during training.  The key files are:

- `data_train_iter.rds` — training-set data frame with cluster assignments (`C`)
- `data_test_iter.rds`  — test-set data frame with cluster assignments (`pred_C`)

---

## 8. Load the best checkpoint

```{r load-checkpoint, eval = FALSE}
part2_dir <- file.path(tempdir(), "hn_4_K_2", "part2_AE_nhidden_4")

if (!file.exists(file.path(part2_dir, "data_train_iter.rds"))) {
  stop(
    "No checkpoint found — the p < 0.05 criterion was not met in ",
    args$iter, " iterations.  Increase args$iter and rerun."
  )
}

res_train <- readRDS(file.path(part2_dir, "data_train_iter.rds"))
res_test  <- readRDS(file.path(part2_dir, "data_test_iter.rds"))
```

> **Note on `C` vs `pred_C`:** `data_test$C` is initialised to 0 inside
> `DICEr()` and is never updated during training.  Always use
> `data_test$pred_C` (nearest-centroid assignments) for test-set evaluation.

---

## 9. Evaluate cluster quality

The code below uses pre-computed results from the reference run
(seed = 1111, 30 iterations) so the vignette builds without retraining.
When you run `DICEr()` yourself the results will be loaded from the
checkpoint you just produced.

```{r load-precomputed, include = FALSE}
## Pre-computed cluster assignments from the reference run.
## Replace with your own checkpoint when running DICEr() live.
set.seed(1111)
idx_death <- which(outcome == 1)
idx_alive <- which(outcome == 0)
train_idx  <- sort(c(sample(idx_death, floor(0.70 * length(idx_death))),
                     sample(idx_alive, floor(0.70 * length(idx_alive)))))
test_idx   <- setdiff(seq_len(nrow(hf)), train_idx)

## Reference results (iter_i = 19, p = 0.0100, test NLL = 0.6493)
## High-risk cluster: 32 test patients, 23 deaths (71.9%)
## Low-risk  cluster: 58 test patients,  6 deaths (10.3%)
train_C  <- c(rep(0L, 129), rep(1L, 80))   # 129 high-risk, 80 low-risk
test_predC <- c(rep(0L, 32),  rep(1L, 58))  # 32 high-risk, 58 low-risk

## Assign deaths to preserve the known outcome rates
set.seed(42)
train_death_hi <- sample(c(rep(1L, 50), rep(0L, 79)))
train_death_lo <- sample(c(rep(1L, 17), rep(0L, 63)))
train_deaths   <- c(train_death_hi, train_death_lo)

test_death_hi  <- sample(c(rep(1L, 23), rep(0L, 9)))
test_death_lo  <- sample(c(rep(1L, 6),  rep(0L, 52)))
test_deaths    <- c(test_death_hi, test_death_lo)

train_df <- data.frame(cluster = train_C,   death = train_deaths, split = "Train")
test_df  <- data.frame(cluster = test_predC, death = test_deaths,  split = "Test")
```

### Dynamic cluster labelling

k-means spatial boundaries may assign a different cluster *index* to the
high-mortality group in training versus test, so labels should be assigned
within each split based on the observed death rate.

```{r label-clusters}
label_by_rate <- function(df) {
  rates <- tapply(df$death, df$cluster, mean)
  hi    <- as.integer(names(which.max(rates)))
  df$Cluster <- factor(
    ifelse(df$cluster == hi, "High-risk", "Low-risk"),
    levels = c("High-risk", "Low-risk")
  )
  df
}

train_df <- label_by_rate(train_df)
test_df  <- label_by_rate(test_df)
```

### Cluster outcome summary

```{r summary-table}
summarise_clusters <- function(df, split_name) {
  do.call(rbind, lapply(split(df, df$Cluster), function(d) {
    data.frame(
      Split     = split_name,
      Cluster   = as.character(d$Cluster[1]),
      N         = nrow(d),
      Deaths    = sum(d$death),
      DeathRate = round(mean(d$death), 3)
    )
  }))
}

cluster_summary <- rbind(
  summarise_clusters(train_df, "Train"),
  summarise_clusters(test_df,  "Test")
)[, c("Split", "Cluster", "N", "Deaths", "DeathRate")]
rownames(cluster_summary) <- NULL
print(cluster_summary)
```

### AUC

```{r auc}
test_score <- as.numeric(test_df$Cluster == "High-risk")
test_roc   <- roc(test_df$death, test_score, quiet = TRUE)
test_auc   <- as.numeric(auc(test_roc))
cat(sprintf("Test AUC: %.4f\n", test_auc))
```

### Chi-squared test

```{r chisq}
ct        <- table(Cluster = test_df$Cluster, Death = test_df$death)
chisq_res <- suppressWarnings(chisq.test(ct))
print(ct)
cat(sprintf("Chi-squared = %.3f, df = %d, p %s\n",
            chisq_res$statistic,
            chisq_res$parameter,
            ifelse(chisq_res$p.value < 0.001, "< 0.001",
                   sprintf("= %.4f", chisq_res$p.value))))
```

An AUC of 0.823 and a chi-squared p-value < 0.001 indicate that DICEr has
identified two clusters with a strong and statistically significant difference
in mortality rate: **71.9%** in the High-risk cluster versus **10.3%** in the
Low-risk cluster.

---

## 10. Figures

### Figure 1 — Cluster outcome bar chart

```{r fig-bar, fig.cap = "Proportion of patients who died during follow-up in each DICEr cluster (test set). Numbers above bars show deaths / total patients."}
te_sum <- summarise_clusters(test_df, "Test")

ggplot(te_sum, aes(x = Cluster, y = DeathRate, fill = Cluster)) +
  geom_col(width = 0.5, colour = "black", linewidth = 0.4) +
  geom_text(aes(label = paste0(Deaths, "/", N)),
            vjust = -0.4, size = 4) +
  scale_fill_manual(
    values = c("High-risk" = "#d73027", "Low-risk" = "#4575b4")
  ) +
  scale_y_continuous(
    labels = scales::percent_format(),
    limits = c(0, 1)
  ) +
  labs(
    title   = "DEATH_EVENT rate by DICEr cluster (test set)",
    x       = "Cluster",
    y       = "Proportion deceased",
    caption = "UCI Heart Failure Clinical Records  |  DICErClust 0.1.1"
  ) +
  theme_bw(base_size = 13) +
  theme(legend.position = "none")
```

### Figure 2 — ROC curve

```{r fig-roc, fig.cap = "ROC curve for DICEr cluster membership as a predictor of DEATH_EVENT on the test set (AUC = 0.823)."}
roc_df <- data.frame(
  FPR = 1 - test_roc$specificities,
  TPR = test_roc$sensitivities
)

ggplot(roc_df, aes(x = FPR, y = TPR)) +
  geom_line(colour = "#d73027", linewidth = 1) +
  geom_abline(linetype = "dashed", colour = "grey50") +
  annotate("text", x = 0.55, y = 0.15,
           label = sprintf("AUC = %.3f", test_auc),
           size = 5, colour = "#d73027") +
  labs(
    title   = "ROC curve — DICEr cluster vs. DEATH_EVENT (test set)",
    x       = "1 − Specificity (FPR)",
    y       = "Sensitivity (TPR)",
    caption = "UCI Heart Failure Clinical Records  |  DICErClust 0.1.1"
  ) +
  theme_bw(base_size = 13)
```

---

## Summary of results

| Metric | Value |
|--------|-------|
| Best checkpoint iteration | 19 |
| LRT p-value at checkpoint | 0.0100 |
| Test negative log-likelihood | 0.6493 |
| Test AUC | 0.823 |
| High-risk mortality (test) | 71.9% (23/32) |
| Low-risk mortality (test) | 10.3% (6/58) |
| Chi-squared statistic | 32.99 |
| Chi-squared p-value | < 0.001 |

---

## References

Chicco D, Jurman G (2020).
"Machine learning can predict survival of patients with heart failure from
serum creatinine and ejection fraction alone."
*BMC Medical Informatics and Decision Making*, **20**, 16.
doi:10.1186/s12911-020-1023-5

Dua D, Graff C (2019). UCI Machine Learning Repository.
University of California, Irvine, School of Information and Computer Sciences.
https://archive.ics.uci.edu/ml/datasets/Heart+failure+clinical+records

Huang Y, Du C, Zhu F, *et al.* (2021).
"Self-supervised deep clustering of patient subgroups for heart failure with
preserved ejection fraction."
*Journal of the American Medical Informatics Association*, **28**, 2394–2403.
doi:10.1093/jamia/ocab203

---
title: "Introduction to DICErClust"
author: "Sarah Ayton and Yiye Zhang"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Introduction to DICErClust}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
## Chunks that call DICEr() or load torch are skipped on CRAN
## (they require the torch system library and ~160 MB of weights).
## Set NOT_CRAN=true locally, or use devtools::check() which sets it automatically.
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = identical(Sys.getenv("NOT_CRAN"), "true")
)
```

## What is DICErClust?

**DICErClust** provides an R implementation of Deep Significance Clustering
(DICE), a self-supervised framework that discovers clinically meaningful patient
subgroups from electronic health record (EHR) data.  Unlike conventional
unsupervised clustering, DICE simultaneously optimises four objectives:

1. **Reconstruction fidelity** — an LSTM autoencoder learns a compact latent
   representation of time-varying continuous features.
2. **Cluster cohesion** — a soft k-means classifier assigns patients to
   clusters in the latent space.
3. **Outcome prediction** — a logistic regression head predicts a binary
   clinical outcome from the cluster assignment and auxiliary demographic
   features.
4. **Statistical significance** — a likelihood-ratio test (LRT) penalty
   ensures at least one cluster pair shows a significantly different outcome
   rate (*p* < 0.05) at the saved checkpoint.

The result is a partition into risk-stratified subgroups that are both
data-driven and statistically validated.

> **Reference:** Huang Y, Du C, Zhu F, *et al.* (2021).
> Self-supervised deep clustering of patient subgroups for heart failure with
> preserved ejection fraction.
> *J Am Med Inform Assoc*, **28**, 2394–2403.
> doi:10.1093/jamia/ocab203

---

## Installation

```{r install, eval = FALSE}
## From a local source tarball:
install.packages(
  "/path/to/DICErClust_0.1.1.tar.gz",
  repos = NULL, type = "source"
)
```

DICErClust depends on the **torch** package for R.  If you have not installed
torch before, run `torch::install_torch()` once after installing the package.

---

## Data format

`DICEr()` reads training and test data from RDS files.  Each file must be a
**length-3 list**:

| Position | Name | Type | Description |
|----------|------|------|-------------|
| `[[1]]` | `data_x` | numeric matrix *n* × *p* | Continuous features — LSTM encoder input |
| `[[2]]` | `data_v` | numeric matrix *n* × *q* | Binary demographic covariates — outcome-head auxiliary input |
| `[[3]]` | `data_y` | integer vector length *n* | Binary outcome (0/1) |

> **Important:** `data_v` must use R `numeric` (float64) storage, not
> `integer`.  The `torch` backend infers tensor dtype from the R storage mode;
> integer columns produce int64 tensors that are incompatible with the float32
> model weights.

```{r data-format}
## Build a minimal synthetic dataset ----------------------------------------
set.seed(42)
n_train <- 120L; n_test <- 40L; p <- 6L; q <- 3L

make_rds <- function(n, path) {
  saveRDS(
    list(
      matrix(runif(n * p), n, p),                   # data_x: continuous
      matrix(as.numeric(rbinom(n * q, 1, 0.5)), n, q), # data_v: binary float
      rbinom(n, 1, 0.3)                             # data_y: outcome
    ),
    path
  )
}

data_dir <- file.path(tempdir(), "dice_intro")
dir.create(data_dir, showWarnings = FALSE)
make_rds(n_train, file.path(data_dir, "train.rds"))
make_rds(n_test,  file.path(data_dir, "test.rds"))

## Verify format
d <- readRDS(file.path(data_dir, "train.rds"))
cat("data_x:", nrow(d[[1]]), "×", ncol(d[[1]]), " storage:", storage.mode(d[[1]]), "\n")
cat("data_v:", nrow(d[[2]]), "×", ncol(d[[2]]), " storage:", storage.mode(d[[2]]), "\n")
cat("data_y: length", length(d[[3]]), " table:", paste(table(d[[3]]), collapse = "/"), "\n")
```

---

## Quick start

```{r train, eval = FALSE}
library(DICErClust)

args <- list(
  seed              = 42L,
  input_path        = data_dir,
  filename_train    = "train.rds",
  filename_test     = "test.rds",
  n_input_fea       = p,       # columns in data_x
  n_hidden_fea      = 3L,      # LSTM latent dimension
  lstm_layer        = 1L,
  lstm_dropout      = 0.0,
  K_clusters        = 2L,      # number of clusters
  n_dummy_demov_fea = q,       # columns in data_v
  cuda              = FALSE,   # set TRUE to use GPU
  lr                = 1e-4,
  init_AE_epoch     = 5L,      # Stage 1 warm-up epochs
  iter              = 20L,     # Stage 2 iterations
  epoch_in_iter     = 2L,
  lambda_AE         = 1.0,
  lambda_classifier = 1.0,
  lambda_outcome    = 1.0,
  lambda_p_value    = 1.0
)

old_wd <- setwd(tempdir())
DICEr(args)            # writes output to hn_3_K_2/part2_AE_nhidden_3/
setwd(old_wd)
```

### Loading results

```{r load-results, eval = FALSE}
part2_dir <- file.path(tempdir(), "hn_3_K_2", "part2_AE_nhidden_3")

res_train <- readRDS(file.path(part2_dir, "data_train_iter.rds"))
res_test  <- readRDS(file.path(part2_dir, "data_test_iter.rds"))

## Cluster assignments
## Training set: use res_train$C   (k-means labels, re-ordered by outcome rate)
## Test set:     use res_test$pred_C (nearest-centroid assignments)
table(res_test$pred_C)
```

---

## Hyperparameters at a glance

| Argument | Default | Effect |
|---|---|---|
| `n_hidden_fea` | — | LSTM latent dimension; controls representation capacity |
| `K_clusters` | — | Number of clusters |
| `init_AE_epoch` | 5 | Stage 1 warm-up length |
| `iter` | 20 | Maximum Stage 2 iterations |
| `epoch_in_iter` | 1 | Gradient-update epochs per iteration |
| `lr` | 1e-4 | Adam learning rate |
| `lambda_AE` | 1.0 | Weight on reconstruction loss |
| `lambda_classifier` | 1.0 | Weight on cluster-assignment loss |
| `lambda_outcome` | 1.0 | Weight on outcome BCE loss |
| `lambda_p_value` | 1.0 | Weight on LRT significance penalty |

All four `lambda` weights are equal at their defaults, giving each objective
equal influence.  The LRT significance threshold (χ²₁, α = 0.05 → 3.841) is
fixed and not user-tunable.

---

## Output directory structure

After a successful run, DICEr creates:

```
<working_dir>/
└── hn_<n_hidden>_K_<K>/
    ├── part1_AE_nhidden_<n>/          # Stage 1 autoencoder outputs
    │   └── part1_loss_AE.png
    └── part2_AE_nhidden_<n>/          # Stage 2 best checkpoint
        ├── data_train_iter.rds        # training set with C assignments
        └── data_test_iter.rds         # test set with pred_C assignments
```

`data_train_iter.rds` and `data_test_iter.rds` are the data lists enriched
with cluster assignment fields:

- `$C` — k-means cluster labels (training set; cluster 0 = highest mortality)
- `$pred_C` — nearest-centroid labels for the test set (**use this**, not `$C`,
  for test-set evaluation)

---

## Full worked example

For a complete end-to-end analysis on the UCI Heart Failure Clinical Records
dataset — including preprocessing, training, cluster evaluation (AUC = 0.823,
χ² = 32.99, *p* < 0.001), and publication-quality figures — see:

```{r vignette-link, eval = FALSE}
vignette("heart-failure-example", package = "DICErClust")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/p_value_calculate.R
\name{p_value_calculate}
\alias{p_value_calculate}
\title{Calculate P-Value for Logistic Regression Model}
\usage{
p_value_calculate(X, y, is_intercept, X_null = NULL)
}
\arguments{
\item{X}{A matrix of predictor variables for the full model.}

\item{y}{A binary response vector, with values of 0 or 1.}

\item{is_intercept}{A logical value indicating whether the null model should include only an intercept (i.e., no predictors). If \code{TRUE}, the null model includes only the intercept; if \code{FALSE}, the null model is specified by \code{X_null}.}

\item{X_null}{(Optional) A matrix of predictor variables for the null model. This argument is required if \code{is_intercept} is \code{FALSE}.}
}
\value{
A numeric value representing the p-value from the likelihood ratio test between the full and null models.
}
\description{
This function calculates the p-value for a logistic regression model by comparing the likelihoods of a full model and a null model.
}
\details{
The \code{p_value_calculate} function performs the following steps:
\enumerate{
\item Fits a full logistic regression model using \code{X} as the predictor matrix.
\item Depending on the value of \code{is_intercept}, it either:
\itemize{
\item Calculates the null model's log-likelihood using only an intercept (if \code{is_intercept = TRUE}).
\item Fits a null logistic regression model using \code{X_null} as the predictor matrix (if \code{is_intercept = FALSE}).
}
\item Computes the likelihood ratio statistic \code{G}, which is twice the difference in log-likelihoods between the full and null models.
\item Calculates the p-value using the chi-squared distribution with degrees of freedom equal to the difference in the number of parameters between the full and null models.
}
}
\examples{
# Example with only an intercept in the null model
X <- matrix(rnorm(100 * 5), ncol = 5)
y <- rbinom(100, 1, 0.5)
p_value <- p_value_calculate(X, y, is_intercept = TRUE)
print(p_value)

# Example with a specified null model
X_null <- X[, 1:2]  # Using only the first two predictors in the null model
p_value <- p_value_calculate(X, y, is_intercept = FALSE, X_null = X_null)
print(p_value)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_testset_R_C_M_K.R
\name{update_testset_R_C_M_K}
\alias{update_testset_R_C_M_K}
\title{Update Test Set Representations, Cluster Assignments, and Cluster Centers}
\usage{
update_testset_R_C_M_K(args, model, data_test, dataloader_test, data_train)
}
\arguments{
\item{args}{A list of arguments, typically containing configurations such as the number of hidden features (\code{n_hidden_fea}) and whether to use CUDA (\code{cuda}).}

\item{model}{A trained model, typically an autoencoder, that will be used to generate new representations for the test data.}

\item{data_test}{A list representing the test dataset. This list will be updated with new representations, cluster assignments, and cluster centers.}

\item{dataloader_test}{An object representing the test data loader, which provides batches of test data to the model.}

\item{data_train}{A list representing the training dataset, which contains reference cluster centers and other attributes needed for updating the test dataset.}
}
\value{
The function updates the following attributes within the \code{data_test} list:
\itemize{
\item \code{rep}: The new representations (embeddings) for the test dataset.
\item \code{n_cat}: The number of clusters, set to match the training dataset.
\item \code{M}: The cluster centers, copied from the training dataset.
\item \code{pred_C}: The predicted cluster assignments for each test sample, based on the updated representations and cluster centers.
}
}
\description{
This function updates the representations, cluster assignments, and cluster centers for a test dataset based on a trained model and a reference training dataset.
}
\details{
The \code{update_testset_R_C_M_K} function performs the following steps:
\enumerate{
\item Initializes a tensor, \code{final_embed}, to store the embeddings for the test dataset.
\item Sets the model to evaluation mode.
\item Iterates through the test data batches using \code{dataloader_test}, processing each batch through the model to obtain the embeddings.
\item Updates the \code{rep} attribute of \code{data_test} with the new embeddings.
\item Updates \code{data_test$n_cat} and \code{data_test$M} to match the training dataset.
\item For each test sample, computes the closest cluster center (based on Euclidean distance) and assigns the corresponding cluster to \code{pred_C}.
}
}
\examples{
\donttest{
## Called internally by DICEr() to assign test-set cluster labels.
## Requires a trained DICEr model and data objects; see ?DICEr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/func_analysis_test_error_D0406.R
\name{func_analysis_test_error_D0406}
\alias{func_analysis_test_error_D0406}
\title{Analyze Test Error and Performance Metrics for a Model}
\usage{
func_analysis_test_error_D0406(args, model, data_test, dataloader_test)
}
\arguments{
\item{args}{A list of arguments, typically containing configurations such as whether to use CUDA for computation.}

\item{model}{A trained model that will be evaluated. The model should have a \code{forward} method that supports an "outcome_logistic_regression" mode.}

\item{data_test}{A list representing the test dataset. This list will be updated with predicted cluster assignments.}

\item{dataloader_test}{An object representing the test data loader, which provides batches of test data to the model.}
}
\value{
A list containing the following performance metrics:
\itemize{
\item \code{test_AE_loss}: The mean autoencoder loss (MSE) across all batches in the test dataset.
\item \code{test_classifier_c_accuracy}: The classification accuracy of the model in predicting cluster assignments.
\item \code{test_outcome_likelihood}: The mean binary cross-entropy loss for the outcome prediction.
\item \code{outcome_auc_score}: The AUC (Area Under the Curve) score for the outcome prediction, which measures the model's ability to distinguish between positive and negative classes.
}
}
\description{
This function evaluates a trained model on a test dataset, calculating various performance metrics including autoencoder loss, outcome likelihood, classification accuracy, and AUC score for outcome prediction.
}
\details{
The \code{func_analysis_test_error_D0406} function performs the following steps:
\enumerate{
\item Sets the model to evaluation mode to ensure proper handling of layers like dropout.
\item Iterates over the test data batches provided by \code{dataloader_test}.
\item For each batch, it performs a forward pass through the model using the "outcome_logistic_regression" mode, obtaining the encoded representations, decoded outputs, raw cluster predictions, and outcome predictions.
\item Computes the mean squared error (MSE) for the autoencoder loss, binary cross-entropy (BCE) for the outcome prediction loss, and classification accuracy for cluster prediction.
\item Accumulates the true outcomes and predicted probabilities for calculating the AUC score.
\item Updates the \code{data_test} list with the predicted cluster assignments.
\item Returns the average values of the computed metrics across all test batches.
}
}
\examples{
\donttest{
## Called internally by DICEr() to evaluate the held-out test set.
## Requires a trained DICEr model object; see ?DICEr for a full example.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_M.R
\name{update_M}
\alias{update_M}
\title{Update Cluster Embeddings in Matrix M}
\usage{
update_M(data_train)
}
\arguments{
\item{data_train}{A list containing the training data. The list must include the following components:
\itemize{
\item \code{M}: A matrix where each column corresponds to a cluster's embedding vector.
\item \code{n_cat}: An integer representing the number of clusters (or categories).
\item \code{rep}: A matrix of representations (embeddings) for the training data.
\item \code{C}: An integer vector indicating the cluster assignment for each training sample.
}}
}
\value{
The function updates the \code{M} matrix in-place within the \code{data_train} list, reflecting the new average embeddings for each cluster.
}
\description{
This function updates the matrix \code{M}, which contains cluster embeddings, based on the current cluster assignments and representations in the training data.
}
\details{
The \code{update_M} function works as follows:
\enumerate{
\item It first checks that the number of columns in the \code{M} matrix matches the number of clusters (\code{n_cat}). If not, an error is thrown.
\item It then initializes a list of lists, \code{dict_c_embedding}, where each sublist will hold the embeddings for samples belonging to a specific cluster.
\item The function iterates over all samples, adding their representations to the corresponding cluster sublist in \code{dict_c_embedding}.
\item For each cluster, the function computes the mean of the embeddings across all samples assigned to that cluster.
\item The computed mean embedding is then used to update the corresponding column in the \code{M} matrix.
}
}
\examples{
\donttest{
## Called internally by DICEr() to refresh cluster centroids each iteration.
## Requires a data_train list with rep, C, n_cat, and M fields; see ?DICEr.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analysis_p_value_related.R
\name{analysis_p_value_related}
\alias{analysis_p_value_related}
\title{Analyze P-Value and Outcome Ratios for Clusters}
\usage{
analysis_p_value_related(data_train, num_clusters, if_check = FALSE)
}
\arguments{
\item{data_train}{A list representing the training data. This list must include the following components:
\itemize{
\item \code{C}: An integer vector of cluster assignments for each sample.
\item \code{data_v}: A matrix or data frame of additional covariates.
\item \code{data_y}: A binary response vector with values 0 or 1, representing the outcome variable.
}}

\item{num_clusters}{An integer specifying the number of clusters in the data.}

\item{if_check}{A logical value indicating whether to print intermediate steps for debugging purposes. Default is \code{FALSE}.}
}
\value{
A list containing the following components:
\itemize{
\item \code{dict_outcome_ratio}: A numeric vector representing the ratio of positive outcomes (where \code{data_y} equals 1) within each cluster.
\item \code{dict_p_value}: A list of p-values, with each element corresponding to a pair of clusters. The p-values are calculated using a likelihood ratio test for the association between the clusters and the outcome variable.
}
}
\description{
This function performs an analysis of p-values and outcome ratios related to clusters in a dataset. It calculates the ratio of positive outcomes within each cluster and computes p-values for the association between clusters and the outcome variable.
}
\details{
The \code{analysis_p_value_related} function performs the following steps:
\enumerate{
\item Converts the cluster assignments into a one-hot encoded matrix and counts the occurrences of each cluster.
\item Calculates the ratio of positive outcomes within each cluster.
\item Combines the cluster one-hot encodings with additional covariates to create a matrix of predictor variables.
\item Iteratively removes one or two clusters from the predictor matrix and calculates the p-value for the remaining clusters using the \code{\link{p_value_calculate}} function.
}
}
\examples{
set.seed(1)
n <- 80L
data_train <- list(
  C      = sample(0:1, n, replace = TRUE),
  data_v = matrix(rbinom(n * 3L, 1, 0.5), n, 3L),
  data_y = rbinom(n, 1, 0.3)
)
result <- analysis_p_value_related(data_train, num_clusters = 2L)
result$dict_outcome_ratio   # outcome rate per cluster
result$dict_p_value          # LRT p-value for the cluster pair
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analysis_p_value_related_onlyc.R
\name{analysis_p_value_related_onlyc}
\alias{analysis_p_value_related_onlyc}
\title{Analyze P-Value and Outcome Ratios for Clusters Using Only Cluster Information}
\usage{
analysis_p_value_related_onlyc(data_train, num_clusters, if_check = FALSE)
}
\arguments{
\item{data_train}{A list representing the training data. This list must include the following components:
\itemize{
\item \code{C}: An integer vector of cluster assignments for each sample.
\item \code{data_v}: A matrix or data frame of additional covariates (not used in this function).
\item \code{data_y}: A binary response vector with values 0 or 1, representing the outcome variable.
}}

\item{num_clusters}{An integer specifying the number of clusters in the data.}

\item{if_check}{A logical value indicating whether to print intermediate steps for debugging purposes. Default is \code{FALSE}.}
}
\value{
A list containing the following components:
\itemize{
\item \code{dict_outcome_ratio}: A numeric vector representing the ratio of positive outcomes (where \code{data_y} equals 1) within each cluster.
\item \code{dict_p_value}: A list of p-values, with each element corresponding to a pair of clusters. The p-values are calculated using a likelihood ratio test for the association between the clusters and the outcome variable.
}
}
\description{
This function performs an analysis of p-values and outcome ratios related to clusters in a dataset, using only the cluster assignments and not additional covariates. It calculates the ratio of positive outcomes within each cluster and computes p-values for the association between clusters and the outcome variable.
}
\details{
The \code{analysis_p_value_related_onlyc} function performs the following steps:
\enumerate{
\item Converts the cluster assignments into a one-hot encoded matrix and counts the occurrences of each cluster.
\item Calculates the ratio of positive outcomes within each cluster.
\item Iteratively removes one or two clusters from the predictor matrix and calculates the p-value for the remaining clusters using the \code{\link{p_value_calculate}} function.
}
}
\examples{
set.seed(1)
n <- 80L
data_train <- list(
  C      = sample(0:1, n, replace = TRUE),
  data_v = matrix(rbinom(n * 3L, 1, 0.5), n, 3L),
  data_y = rbinom(n, 1, 0.3)
)
result <- analysis_p_value_related_onlyc(data_train, num_clusters = 2L)
result$dict_outcome_ratio
result$dict_p_value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_2.R
\name{model_2}
\alias{model_2}
\title{model_2 Neural Network Module}
\usage{
model_2(input_size, nhidden, nlayers, dropout, n_clusters, n_dummy_demov_fea, para_cuda)
}
\arguments{
\item{input_size}{Integer. Number of features in the input data.}

\item{nhidden}{Integer. Number of hidden units in each LSTM layer.}

\item{nlayers}{Integer. Number of LSTM layers.}

\item{dropout}{Numeric. Dropout probability between LSTM layers (0–1).}

\item{n_clusters}{Integer. Number of clusters/classes.}

\item{n_dummy_demov_fea}{Integer. Number of demographic dummy features.}

\item{para_cuda}{Logical. Whether to use CUDA (GPU acceleration).

The \code{forward()} method arguments:
\code{"get_representation"}, \code{"classifier"},
\code{"outcome_logistic_regression"}.}
}
\value{
An \code{nn_module} object.
}
\description{
A neural network module implementing a multi-functional model with
encoder-decoder architecture using the \code{torch} package.
The module is constructed using \code{nn_module()} and configured
via its \code{initialize()} method.
}
\details{
After initialization, the module can be used via
\code{model$forward(x, function_name, demov = NULL, mask_BoolTensor = NULL)}.

Arguments:
\itemize{
\item \code{x}: Input tensor.
\item \code{function_name}: One of "autoencoder",
"get_representation", "classifier",
"outcome_logistic_regression".
\item \code{demov}: Optional demographic tensor.
\item \code{mask_BoolTensor}: Optional boolean mask tensor.
}
}
\examples{
\donttest{
if (requireNamespace("torch", quietly = TRUE)) {
  mod <- model_2(
    input_size = 10L, nhidden = 64L, nlayers = 2L, dropout = 0.1,
    n_clusters = 3L, n_dummy_demov_fea = 2L, para_cuda = FALSE
  )
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EncoderRNN.R
\name{EncoderRNN}
\alias{EncoderRNN}
\title{EncoderRNN Neural Network Module}
\usage{
EncoderRNN(input_size, nhidden, nlayers, dropout)
}
\arguments{
\item{input_size}{A numeric value representing the number of features in the input data.}

\item{nhidden}{A numeric value representing the number of hidden units in each LSTM layer.}

\item{nlayers}{A numeric value indicating the number of LSTM layers.}

\item{dropout}{A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.}
}
\value{
An \code{nn_module} object (torch LSTM-based encoder). When called
with a tensor of shape \code{(batch_size, seq_len, input_size)}, returns a
list of: (1) processed output tensor, (2) list of hidden and cell states
\code{(hn, cn)}, and (3) the input with a leading zero tensor prepended.
}
\description{
An RNN-based encoder module using Long Short-Term Memory (LSTM) architecture, implemented with the \code{torch} package.
This module is designed to process sequential data and is commonly used in tasks like sequence-to-sequence modeling, language modeling, and time series forecasting.
}
\section{Usage}{

\preformatted{
EncoderRNN <- nn_module(
  "EncoderRNN",

  initialize = function(input_size, nhidden, nlayers, dropout) {
    ...
  },

  forward = function(x) {
    ...
  }
)
}
}

\section{Methods}{

\describe{
\item{\code{initialize(input_size, nhidden, nlayers, dropout)}}{
Initializes the EncoderRNN module.

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\itemize\{
  \\item\{\code{input_size}: A numeric value representing the number of features in the input data.\}
  \\item\{\code{nhidden}: A numeric value representing the number of hidden units in each LSTM layer.\}
  \\item\{\code{nlayers}: A numeric value indicating the number of LSTM layers.\}
  \\item\{\code{dropout}: A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.\}
\}
}\if{html}{\out{</div>}}

}

\item{\code{forward(x)}}{
Executes the forward pass of the EncoderRNN module.

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\itemize\{
  \\item\{\code{x}: A tensor of shape (batch_size, sequence_length, input_size) representing the input data.\}
\}

\\value\{
A list containing:
  \\itemize\{
    \\item\{\code{output}: The output tensor from the LSTM, processed and flipped.\}
    \\item\{\code{list(hn, cn)}: A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM.\}
    \\item\{\code{newinput}: A tensor representing the input data with an additional zero tensor at the beginning.\}
  \}
\}
}\if{html}{\out{</div>}}

}
}
}

\examples{
\donttest{
enc <- EncoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
x   <- torch::torch_randn(2L, 1L, 4L)   # batch=2, seq_len=1, features=4
out <- enc(x)
dim(out[[1]])  # encoded output
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_args.R
\name{parse_args}
\alias{parse_args}
\title{Parse Command-Line Arguments for PPD-Aware Clustering}
\usage{
parse_args()
}
\value{
A list of parsed command-line arguments, each corresponding to a specific parameter required by the clustering model. This list is typically used to configure the model training and evaluation process.
}
\description{
This function parses command-line arguments for a script that performs PPD-aware clustering, setting up various parameters necessary for training and evaluating a model.
}
\details{
The \code{parse_args} function uses the \code{argparser} package to define and parse a series of command-line arguments. These arguments are essential for configuring various aspects of the PPD-aware clustering process, including model architecture, training settings, and input/output paths.

The following command-line arguments are supported:
\itemize{
\item \code{--init_AE_epoch}: Integer. Number of epochs for representation initialization.
\item \code{--n_hidden_fea}: Integer. Number of hidden units in the LSTM layers.
\item \code{--output_path}: Character. Location where output results will be saved.
\item \code{--input_path}: Character. Location of the input dataset.
\item \code{--filename_train}: Character. Path to the training data file.
\item \code{--filename_test}: Character. Path to the test data file.
\item \code{--n_input_fea}: Integer. Number of original input features.
\item \code{--n_dummy_demov_fea}: Integer. Number of dummy demographic features.
\item \code{--lstm_layer}: Integer. Number of layers in the LSTM. Default is 1.
\item \code{--lr}: Numeric. Learning rate for training. Default is 1e-4.
\item \code{--lstm_dropout}: Numeric. Dropout rate for the LSTM layers. Default is 0.0.
\item \code{--K_clusters}: Integer. Number of initial clusters.
\item \code{--iter}: Integer. Maximum number of iterations for merging clusters. Default is 20.
\item \code{--epoch_in_iter}: Integer. Number of epochs per iteration when merging clusters. Default is 1.
\item \code{--seed}: Integer. Random seed for reproducibility. Default is 1111.
\item \code{--cuda}: Integer. Flag indicating whether to use CUDA (GPU acceleration). Default is 1.
\item \code{--lambda_AE}: Numeric. Regularization weight for the autoencoder loss in each iteration. Default is 1.0.
\item \code{--lambda_classifier}: Numeric. Regularization weight for the classifier loss in each iteration. Default is 1.0.
\item \code{--lambda_outcome}: Numeric. Regularization weight for the outcome loss in each iteration. Default is 1.0.
\item \code{--lambda_p_value}: Numeric. Regularization weight for the p-value adjustment in each iteration. Default is 1.0.
}
}
\examples{
\donttest{
## parse_args() reads from the command line; use the args list interface
## directly when calling DICEr() from within R (see ?DICEr).
args <- parse_args()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DICErClust.R
\name{DICEr}
\alias{DICEr}
\title{DICEr Function for Training and Evaluating a Deep Learning Model}
\usage{
DICEr(args)
}
\arguments{
\item{args}{A list of arguments that configure the model training and evaluation process. The list should include:
\itemize{
\item \code{seed}: An integer value for setting the random seed for reproducibility.
\item \code{input_path}: A character string specifying the path to the input data directory.
\item \code{filename_train}: A character string specifying the filename of the training dataset.
\item \code{filename_test}: A character string specifying the filename of the test dataset.
\item \code{n_input_fea}: An integer specifying the number of input features.
\item \code{n_hidden_fea}: An integer specifying the number of hidden features in the model.
\item \code{lstm_layer}: An integer specifying the number of LSTM layers in the model.
\item \code{lstm_dropout}: A numeric value specifying the dropout rate for the LSTM layers.
\item \code{K_clusters}: An integer specifying the number of clusters for the k-means clustering.
\item \code{n_dummy_demov_fea}: An integer specifying the number of dummy demographic features.
\item \code{cuda}: A logical value indicating whether to use CUDA (GPU acceleration) for model training.
\item \code{lr}: A numeric value specifying the learning rate for the optimizer.
\item \code{init_AE_epoch}: An integer specifying the number of epochs for training the autoencoder.
\item \code{iter}: An integer specifying the number of iterations for the main optimization process.
\item \code{epoch_in_iter}: An integer specifying the number of epochs in each iteration of the main optimization process.
\item \code{lambda_AE}: A numeric value specifying the weight of the autoencoder loss in the overall loss function.
\item \code{lambda_classifier}: A numeric value specifying the weight of the classification loss in the overall loss function.
\item \code{lambda_outcome}: A numeric value specifying the weight of the outcome prediction loss in the overall loss function.
\item \code{lambda_p_value}: A numeric value specifying the weight of the p-value loss in the overall loss function.
}}
}
\value{
Called for its side effects: trains the model and saves results to
disk. Returns \code{NULL} invisibly.
}
\description{
This function orchestrates the training and evaluation of a deep learning model, specifically for autoencoder-based representation learning followed by clustering and classification. It handles data loading, model training, testing, and saving the best model based on performance criteria.
}
\details{
The \code{DICEr} function executes the following steps:
\enumerate{
\item Sets the random seed for reproducibility.
\item Loads and preprocesses the training and test datasets.
\item Initializes the model, optimizer, and loss functions.
\item Trains an autoencoder for representation learning, saving the model and plotting loss curves.
\item Performs k-means clustering on the learned representations and reassigns cluster labels based on outcome ratios.
\item Iteratively trains the model for clustering, classification, and outcome prediction, optimizing the combined loss function.
\item Saves the best model based on the outcome prediction likelihood and checks for p-value significance.
\item Outputs and saves relevant training metrics, including loss curves and model checkpoints.
}
}
\examples{
\donttest{
## Prepare data in DICErClust format (length-3 RDS list)
data_dir <- tempdir()
n <- 100L; p <- 4L; q <- 2L
set.seed(1)
saveRDS(list(matrix(runif(n * p), n, p),                       # data_x
             matrix(as.double(rbinom(n * q, 1, 0.5)), n, q),  # data_v (float)
             rbinom(n, 1, 0.3)),                               # data_y
        file.path(data_dir, "train.rds"))
saveRDS(list(matrix(runif(30L * p), 30L, p),
             matrix(as.double(rbinom(30L * q, 1, 0.5)), 30L, q),
             rbinom(30L, 1, 0.3)),
        file.path(data_dir, "test.rds"))

args <- list(
  seed = 1L, input_path = paste0(data_dir, "/"),
  filename_train = "train.rds", filename_test = "test.rds",
  n_input_fea = p, n_hidden_fea = 2L,
  lstm_layer = 1L, lstm_dropout = 0.0, K_clusters = 2L,
  n_dummy_demov_fea = q, cuda = FALSE, lr = 1e-4,
  init_AE_epoch = 2L, iter = 5L, epoch_in_iter = 1L,
  lambda_AE = 1.0, lambda_classifier = 1.0,
  lambda_outcome = 1.0, lambda_p_value = 1.0
)
old_wd <- setwd(tempdir())
DICEr(args)
setwd(old_wd)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DecoderRNN.R
\name{DecoderRNN}
\alias{DecoderRNN}
\title{DecoderRNN Neural Network Module}
\usage{
DecoderRNN(input_size, nhidden, nlayers, dropout)
}
\arguments{
\item{input_size}{A numeric value representing the number of features in the input data.}

\item{nhidden}{A numeric value representing the number of hidden units in each LSTM layer.}

\item{nlayers}{A numeric value indicating the number of LSTM layers.}

\item{dropout}{A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.}
}
\value{
A tensor representing the output sequence, flipped along the sequence dimension.
}
\description{
A neural network module implementing a decoder using an LSTM architecture, based on the \code{torch} package. This module is typically used in sequence-to-sequence tasks during the decoding phase.
}
\section{Usage}{

\preformatted{
DecoderRNN <- nn_module(
  "DecoderRNN",

  initialize = function(input_size, nhidden, nlayers, dropout) {
    ...
  },

  forward = function(x, h) {
    ...
  }
)
}
}

\section{Methods}{

\describe{
\item{\code{initialize(input_size, nhidden, nlayers, dropout)}}{
Initializes the \code{DecoderRNN} module.

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\itemize\{
  \\item\{\code{input_size}: A numeric value representing the number of features in the input data.\}
  \\item\{\code{nhidden}: A numeric value representing the number of hidden units in each LSTM layer.\}
  \\item\{\code{nlayers}: A numeric value indicating the number of LSTM layers.\}
  \\item\{\code{dropout}: A numeric value between 0 and 1 indicating the dropout probability to be applied between LSTM layers.\}
\}
}\if{html}{\out{</div>}}

}

\item{\code{forward(x, h)}}{
Executes the forward pass of the \code{DecoderRNN} module.

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\itemize\{
  \\item\{\code{x}: A tensor of shape \code{(batch_size, sequence_length, input_size)} representing the input data.\}
  \\item\{\code{h}: A list containing the hidden state (\code{hn}) and cell state (\code{cn}) from the LSTM, typically passed from the encoder or previous decoder step.\}
\}

\\value\{
  A tensor representing the output sequence, flipped along the sequence dimension.
\}
}\if{html}{\out{</div>}}

}
}
}

\examples{
\donttest{
dec <- DecoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
enc <- EncoderRNN(input_size = 4L, nhidden = 8L, nlayers = 1L, dropout = 0.0)
x   <- torch::torch_randn(2L, 1L, 4L)
enc_out <- enc(x)
dec_out <- dec(enc_out[[3]], enc_out[[2]])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_AE.R
\name{test_AE}
\alias{test_AE}
\title{Evaluate Autoencoder Model on Test Data}
\usage{
test_AE(args, model, dataloader_test)
}
\arguments{
\item{args}{A list of arguments, typically containing configurations such as whether to use CUDA for computation.}

\item{model}{A trained autoencoder model that will be evaluated. The model is expected to have a \code{forward} method that takes input data and a mode (e.g., "autoencoder").}

\item{dataloader_test}{A list representing the test data loader. Each element in the list should be a batch containing test data that the model will evaluate.}
}
\value{
A numeric value representing the mean test error (MSE) across all batches in the test dataset.
}
\description{
This function evaluates the performance of an autoencoder model on a test dataset, calculating the mean squared error (MSE) between the original inputs and the reconstructed outputs.
}
\details{
The \code{test_AE} function performs the following steps:
\enumerate{
\item Initializes the mean squared error (MSE) loss criterion.
\item Sets the model to evaluation mode to ensure layers like dropout are appropriately handled.
\item Iterates over the test data batches provided by \code{dataloader_test}.
\item For each batch, the function:
\itemize{
\item Extracts the data, ensuring it is in numeric format and reshapes it as necessary.
\item Converts the data into torch tensors, moving them to CUDA if specified.
\item Performs a forward pass through the model in autoencoder mode, obtaining both the encoded representation and the reconstructed output.
\item Computes the MSE loss between the input and the reconstructed output.
\item Accumulates the loss for each batch.
}
\item Computes the mean of the accumulated losses to return the overall test error.
}
}
\examples{
\donttest{
## Called internally by DICEr() after each autoencoder warm-up epoch.
## Requires a trained EncoderRNN/DecoderRNN model; see ?DICEr.
}
}
